"""
hydrofunctions.usgs_rdb
~~~~~~~~~~~~~~~~~~~~~~~

This module is for working with the various USGS dataservices that use the rdb
text format. These include the statistics service, the field measurements
service, the rating curve service, and the peak discharge service.
"""
import pandas as pd
import requests
from io import StringIO
from IPython.core import display
from . import exceptions


class hydroRDB:
    """A class for holding the information from USGS rdb files.

    Args:
        header (str):
            A multi-line string from the header of the rdb file. The header
            often contain important metadata and user warnings.
        table (pandas dataframe):
            This is a dataframe made from the rdb file.
        columns (str):
            A string from the rdb file that lists the column names.
        dtypes (str):
            A string from the rdb file that gives the data type and length of
            each column.
        rdb (str):
            The complete original text of the rdb file.

    **Properties:**
        **header** (str):
            A multi-line string from the header of the rdb file. The header
            often contain important metadata and user warnings.
        **table** (pandas dataframe):
            This is a dataframe made from the rdb file.
        **columns** (str):
            A string from the rdb file that lists the column names.
        **dtypes** (str):
            A string from the rdb file that gives the data type and length of
            each column.
        **rdb** (str):
            The original, unparsed rdb file as returned by the USGS.

        You can also access the header and the dataframe as a named tuple::

            hydroRDB(header=<a multi-line string>, table=<pandas dataframe>)

    Note:
        - The args to create this object are supplied by hf.read_rdb().
        - The hydroRDB object is returned from several functions that request\
        RDB files from a USGS data service, including: peaks(), field_meas(),\
        rating_curve(), stats(), site_file(), and data_catalog().
        - You can read more about the RDB format here: https://pubs.usgs.gov/of/2003/ofr03123/6.4rdb_format.pdf
    """

    def __init__(self, header, table, columns, dtypes, rdb_str):
        self.header = header
        self.table = table
        self.columns = columns
        self.dtypes = dtypes
        self.rdb = rdb_str

    def __iter__(self):
        return iter((self.header, self.table))

    def __repr__(self):
        return f"hydroRDB(header={self.header},\ntable={self.table}"

    def _repr_html_(self):
        html_header = "<p>" + self.header.replace("\n", "<br />") + "</p>"
        # return html_header + self.df._repr_html_()
        return f"<p>hydroRDB(header=<br />{html_header}</p><p>table=<br />{self.table._repr_html_()})</p>"


def get_usgs_RDB_service(url, headers=None, params=None):
    """Request data from a USGS dataservice and handle errors.

    Args:
        url (str):
            a string used by Requests as the base URL.
        header (dict):
            a dict of parameters used to request the data.
        params (dict):
            a dict of parameters used to modify the url of a REST service.
    Returns:
        A Requests response object.
    Raises:
        This function will raise an exception for any non-200 status code, and\
        in cases where the USGS service returns anything that is not obviously\
        an RDB file. If an exception is raised, then an attempt will be made to\
        display the error page which the USGS sometimes sends back to the user.
    """
    response = requests.get(url, headers=headers, params=params)
    if response.status_code == 200:
        if response.text[0] == "#":
            # Everything seems good; they apparently returned an RDB file.
            return response
        else:
            print(
                "The USGS has apparently not returned any data. Check the "
                "following message for further information for why this "
                "request failed. One possibility is that your site number "
                "is incorrect."
            )
            display.display(display.HTML(response.text))
            raise exceptions.HydroNoDataError(
                "The USGS did not return a valid RDB file for this request."
            )
    else:
        # response.status_code != 200:
        print(f"The USGS has returned an error code of {response.status_code}")
        # If this code is being run inside of a notebook, the USGS error page
        # will be displayed.
        display.display(display.HTML(response.text))
        # raise an exception
        response.raise_for_status()
        # or raise some sort of Hydro http error based on requests http error.
    return response


def read_rdb(text):
    """Read strings that are in rdb format.

    Args:
        text (str):
            A long string containing the contents of a rdb file. A common way
            to obtain these would be from the .text property of a requests
            response, as in the example usage below.

    Returns:
        header (multi-line string):
            Every commented line at the top of the rdb file is marked with a
            '#' symbol. Each of these lines is stored in this output.
        outputDF (pandas.DataFrame):
            A dataframe containing the information in the rdb file. `site_no`
            and `parameter_cd` are interpreted as a string, but every other number
            is interpreted as a float or int; missing values as an np.nan;
            strings for everything else.
        columns (list of strings):
            The column names, taken from the rdb header row.
        dtypes (list of strings):
            The second header row from the rdb file. These mostly tell the
            column width, and typically record everything as string data ('s')
            type. The exception to this are dates, which are listed with a 'd'.
    """
    try:
        headerlines = []
        datalines = []
        count = 0
        for line in text.splitlines():
            if line[0] == "#":
                headerlines.append(line)
            elif count == 0:
                columns = line.split()
                count += 1
            elif count == 1:
                dtypes = line.split()
                count += 1
            else:
                datalines.append(line)
        data = "\n".join(datalines)
        header = "\n".join(headerlines)

        outputDF = pd.read_csv(
            StringIO(data),
            sep="\t",
            comment="#",
            header=None,
            names=columns,
            dtype={"site_no": str, "parameter_cd": str},
            # When converted like this, poorly-formed dates will be saved as strings
            parse_dates=True,
            # If dates are converted like this, then poorly formed dates will stop the process
            # converters={"peak_dt": pd.to_datetime},
        )
        # Another approach would be to convert date columns later, and catch errors
        # try:
        #   outputDF.peak_dt = pd.to_datetime(outputDF.peak_dt)
        # except ValueError as err:
        #   print(f"Unable to parse date. reason: '{str(err)}'.")

    except:
        print(
            "There appears to be an error processing the file that the USGS "
            "returned. This sometimes occurs if you entered the wrong site "
            "number. We were expecting an RDB file, but we received the "
            f"following instead:\n{text}"
        )
        raise
    # outputDF.outputDF.filter(like='_cd').columns
    # TODO: code columns ('*._cd') should be interpreted as strings.
    # TODO: date columns ('*_dt') should be converted to dates.

    return header, outputDF, columns, dtypes


def site_file(site, verbose=True):
    """Load USGS site file into a Pandas dataframe.

    Args:
        site (str):
            The gauge ID number for the site.
        verbose (bool):
            If True (default), will print confirmation messages with the url before and
            after the request.

    Returns:
        a hydroRDB object or tuple consisting of the header and a pandas
        dataframe. The dataframe will have one row for every site requested; for each
        site it will provide detailed site characteristics such as watershed area, 
        drainage basin HUC code, site latitude, longitude, altitude, and datum; the date
        the site was established, hole depth for wells, and other information. All of
        the columns are listed in the header; for more information, visit: 
        http://waterservices.usgs.gov/rest/Site-Service.html

        For information on the data collected at this site (including the start and stop
        dates for data collection), use the 'data_catalog' function. 

    **Example:**

        >>> test = site_file('01542500')

        >>> test
        hydroRDB(header=<a multi-line string of the header>,
                 table=<a Pandas dataframe>)

    You can also access the header, dataframe, column names, and data types
    through the associated properties `header`, `table`, `columns`, `dtypes`::

        >>> test.table
        <a Pandas dataframe>

    """
    url = (
        "https://waterservices.usgs.gov/nwis/site/?format=rdb&sites="
        + site
        + "&siteOutput=expanded&siteStatus=all"
    )
    headers = {"Accept-encoding": "gzip"}

    if verbose:
        print(f"Retrieving the site file for site #{site} from {url}", end="\r")
    response = get_usgs_RDB_service(url, headers)
    if verbose:
        print(f"Retrieved the site file for site #{site} from {url}")
    (
        header,
        outputDF,
        columns,
        dtype,
    ) = read_rdb(response.text)

    return hydroRDB(header, outputDF, columns, dtype, response.text)


def data_catalog(site, verbose=True):
    """Load a history of the data collected at a site into a Pandas dataframe.

    Args:
        site (str):
            The gauge ID number for the site.
        verbose (bool):
            If True (default), will print confirmation messages with the url before and
            after the request.

    Returns:
        a hydroRDB object or tuple consisting of the header and a pandas
        dataframe.  The dataframe will have one row for every type of data collected at
        each site requested; for each data parameter it will provide information 
        including: parameter code, date of first observation, date of last observation,
        and total number of observations. A full description of the data catalog is
        given in the header; more information is available at: 
        http://waterservices.usgs.gov/rest/Site-Service.html

        For information about the site itself, including watershed area and HUC code, 
        use the 'site_file' function.

    **Example:**

        >>> test = data_catalog('01542500')
        >>> test
        hydroRDB(header=<a mulit-line string of the header>,
                 table=<a Pandas dataframe>)

    You can also access the header, dataframe, column names, and data types
    through the associated properties `header`, `table`, `columns`, `dtypes`::

        >>> test.table
        <a Pandas dataframe>

    """
    url = (
        "https://waterservices.usgs.gov/nwis/site/?format=rdb&sites="
        + site
        + "&seriesCatalogOutput=true&siteStatus=all"
    )
    headers = {"Accept-encoding": "gzip"}

    if verbose:
        print(f"Retrieving the data catalog for site #{site} from {url}", end="\r")
    response = get_usgs_RDB_service(url, headers)
    if verbose:
        print(f"Retrieved the data catalog for site #{site} from {url}")
    (
        header,
        outputDF,
        columns,
        dtype,
    ) = read_rdb(response.text)
    return hydroRDB(header, outputDF, columns, dtype, response.text)


def field_meas(site, verbose=True):
    """Load USGS field measurements of stream discharge into a Pandas dataframe.

    Args:
        site (str):
            The gauge ID number for the site.
        verbose (bool):
            If True (default), will print confirmation messages with the url before and
            after the request.

    Returns:
        a hydroRDB object or tuple consisting of the header and a pandas
        dataframe. Each row of the table represents an observation on a given date of
        river conditions at the gauge by USGS personnel. Values are stored in
        columns, and include the measured stream discharge, channel width,
        channel area, depth, and velocity.

    **Example:**

        >>> test = field_meas('01542500')
        >>> test
        hydroRDB(header=<a mulit-line string of the header>,
                 table=<a Pandas dataframe>)

    You can also access the header, dataframe, column names, and data types
    through the associated properties `header`, `table`, `columns`, `dtypes`::

        >>> test.table
        <a Pandas dataframe>

    **Discussion:**
        The USGS operates over 8,000 stream gages around the United States and
        territories. Each of these sensors records the depth, or 'stage' of the
        water. In order to translate this stage data into stream discharge, the
        USGS staff creates an empirical relationship called a 'rating curve'
        between the river stage and stream discharge. To construct this curve,
        the USGS personnel visit all of the gage every one to eight weeks, and
        measure the stage and the discharge of the river manually.

        The ``field_meas()`` function returns all of the field-collected data for
        this site. The USGS uses these data to create the rating curve. You can use
        these data to see how the site has changed over time, or to
        read the notes about local conditions.

        The ``rating_curve()`` function returns the most recent 'expanded shift-
        adjusted' rating curve constructed for this site. This is the current official
        rating curve.

        To plot a rating curve from the field measurements, use::

            >>> header, data = hf.field_meas('01581830')

            >>> data.plot(x='gage_height_va', y='discharge_va', kind='scatter')

        Rating curves are typically plotted with the indepedent variable,
        gage_height, plotted on the Y axis.
    """
    url = (
        "https://waterdata.usgs.gov/nwis/measurements?site_no="
        + site
        + "&agency_cd=USGS&format=rdb_expanded"
    )
    headers = {"Accept-encoding": "gzip"}

    if verbose:
        print(
            f"Retrieving the field measurements for site #{site} from {url}", end="\r"
        )
    response = get_usgs_RDB_service(url, headers)
    if verbose:
        print(f"Retrieved the field measurements for site #{site} from {url}")
    # It may be desireable to keep the original na_values, like 'unkn' for many
    # of the columns. However, it is still a good idea to replace for the gage
    # depth and discharge values, since these variables get used in plotting
    # functions.
    (
        header,
        outputDF,
        columns,
        dtype,
    ) = read_rdb(response.text)

    try:
        outputDF.measurement_dt = pd.to_datetime(outputDF.measurement_dt)
    except ValueError as err:
        print(
            f"Unable to parse the measurement_dt field as a date. reason: '{str(err)}'."
        )

    # An attempt to use the tz_cd column to make measurement_dt timezone aware.
    # outputDF.tz_cd.replace({np.nan: 'UTC'}, inplace=True)
    # def f(x, y):
    #    return x.tz_localize(y)
    # outputDF['datetime'] = outputDF[['measurement_dt', 'tz_cd']].apply(lambda x: f(*x), axis=1)

    outputDF.set_index("measurement_dt", inplace=True)
    return hydroRDB(header, outputDF, columns, dtype, response.text)


def peaks(site, verbose=True):
    """Return a series of annual peak discharges.

    Args:
        site(str):
            The gauge ID number for the site.
        verbose (bool):
            If True (default), will print confirmation messages with the url before and
            after the request.

    Returns:
        a hydroRDB object or tuple consisting of the header and a table. The header
        is a multi-line string of metadata supplied by the USGS with the data series.
        The table is a dataframe containing the annual peak discharge series. You can
        use these data to conduct a flood frequency analysis.

    **Example:**

        >>> test = hf.peaks('01542500')
        >>> test
        hydroRDB(header=<a mulit-line string of the header>,
                 table=<a Pandas dataframe>)

    You can also access the header, dataframe, column names, and data types
    through the associated properties `header`, `table`, `columns`, `dtypes`::

        >>> test.table
        <a Pandas dataframe>

    """
    url = (
        "https://nwis.waterdata.usgs.gov/nwis/peak?site_no="
        + site
        + "&agency_cd=USGS&format=rdb"
    )

    headers = {"Accept-encoding": "gzip"}

    if verbose:
        print(
            f"Retrieving the annual peak discharges for site #{site} from {url}",
            end="\r",
        )
    response = get_usgs_RDB_service(url, headers)
    if verbose:
        print(f"Retrieved the annual peak discharges for site #{site} from {url}")
    header, outputDF, columns, dtype = read_rdb(response.text)
    try:
        outputDF.peak_dt = pd.to_datetime(outputDF.peak_dt)
    except ValueError as err:
        print(f"Unable to parse the peak_dt field as a date. Reason: '{str(err)}'.")

    # peak_date might be a string, or it might be a datetime. Both work as an index
    outputDF.set_index("peak_dt", inplace=True)
    return hydroRDB(header, outputDF, columns, dtype, response.text)


def rating_curve(site, verbose=True):
    """Return the most recent USGS expanded-shift-adjusted rating curve for a
    given stream gage into a dataframe.

    Args:
        site (str):
            The gage ID number for the site.
        verbose (bool):
            If True (default), will print confirmation messages with the url before and
            after the request.

    Returns:
        a hydroRDB object or tuple consisting of the header and a table. The header
        is a multi-line string of metadata supplied by the USGS with the data series.
        The table is a dataframe containing the latest official rating curve for the
        site.

    **Example:**

        >>> test = rating_curve('01542500')
        >>> test
        hydroRDB(header=<a mulit-line string of the header>,
                 table=<a Pandas dataframe>)

    You can also access the header, dataframe, column names, and data types
    through the associated properties `header`, `table`, `columns`, `dtypes`::

        >>> test.table
        <a Pandas dataframe>

    **Discussion:**
        The USGS operates over 8,000 stream gauges around the United States and
        territories. Each of these sensors records the depth, or 'stage' of the
        water. In order to translate this stage data into stream discharge, the
        USGS staff creates an empirical relationship called a 'rating curve'
        between the river stage and stream discharge.

        See ``hf.field_meas()`` to access the field data used to construct the
        rating curve.

        **Note:** Rating curves change over time.
    """
    url = (
        "https://waterdata.usgs.gov/nwisweb/data/ratings/exsa/USGS."
        + site
        + ".exsa.rdb"
    )
    headers = {"Accept-encoding": "gzip"}
    if verbose:
        print(f"Retrieving the rating curve for site #{site} from {url}", end="\r")
    response = get_usgs_RDB_service(url, headers)
    if verbose:
        print(f"Retrieved the rating curve for site #{site} from {url}")
    header, outputDF, columns, dtype = read_rdb(response.text)
    outputDF.columns = ["stage", "shift", "discharge", "stor"]
    """
    outputDF = pd.read_csv(StringIO(response.text),
                     sep='\t',
                     comment='#',
                     header=1,
                     names=['stage', 'shift', 'discharge', 'stor'],
                     skiprows=2
                     )
    """
    return hydroRDB(header, outputDF, columns, dtype, response.text)


def stats(site, statReportType="daily", verbose=True, **kwargs):
    """Return statistics from the USGS Stats Service as a dataframe.

    Args:
        site (str):
            The gage ID number for the site, or a series of gage IDs separated
            by commas, like this: '01546500,01548000'.

        statReportType ('daily'|'monthly'|'annual'):
            There are three different types of report that you can request.

            - 'daily' (default): calculate statistics for each of 365 days.
            - 'monthly': calculate statistics for each of the twelve months.
            - 'annual': calculate annual statistics for each year since the start of the record.
        verbose (bool):
            If True (default), will print confirmation messages with the url before and
            after the request.

    Returns:
        a hydroRDB object or tuple consisting of the header and a table. The header
        is a multi-line string of metadata supplied by the USGS with the data series.
        The table is a dataframe containing the latest official statistics for the
        site.

    Raises:
        HTTPError
            when a non-200 http status code is returned.

    **Example:**

        >>> test = stats('01542500', 'monthly')
        >>> test
        hydroRDB(header=<a mulit-line string of the header>,
                 table=<a Pandas dataframe>)

    You can also access the header, dataframe, column names, and data types
    through the associated properties `header`, `table`, `columns`, `dtypes`::

        >>> test.table
        <a Pandas dataframe>

    Note:
        This function is based on the USGS statistics service, described here:
        https://waterservices.usgs.gov/rest/Statistics-Service.html

        The USGS Statistics Service allows you to specify a wide array of
        additional parameters in your request. You can provide these parameters
        as keyword arguments, like in this example::

            >>> hf.stats('01452500', parameterCD='00060')

        This will only request
        statistics for discharge, which is specified with the '00060'
        parameter code.

        Additional useful parameters include:

            - `parameterCD='00060,00065'` Limit the request for statistics to
              only one parameter or to a list of parameters. The default behavior
              is to provide statistics for every parameter that has been measured
              at this site. In this example, statistics for discharge ('00060')
              and stage ('00065') are requested.

            - `statYearType='water'` Calculate annual statistics based on the
              water year, which runs from October 1st to September 31st. This
              parameter is only for use with annual reports. If not specified,
              the default behavior will use calendar years for reporting.

            - `missingData='on'`  Calculate statistics even when there are some
              missing values. If not specified, the default behavior is to drop
              years that have fewer than 365 values from annual reports, and to
              drop months that have fewer than 30 values in monthly reports. The
              number of values used to calculate a statistic is reported in the
              'count_nu' column.

            - You can read about other useful parameters here: https://waterservices.usgs.gov/rest/Statistics-Service.html#statistical_Controls

    """
    url = "https://waterservices.usgs.gov/nwis/stat/"

    headers = {
        "Accept-encoding": "gzip",
    }
    # Set default values for parameters that are too obscure to put into call
    # signature.
    params = {
        "statReportType": statReportType,
        "statType": "all",
        "sites": site,
        "format": "rdb",
    }
    # Overwrite defaults if they are specified.
    params.update(kwargs)

    if verbose:
        print(
            f"Retrieving {params['statReportType']} statistics for site #{params['sites']} from {url}",
            end="\r",
        )
    response = get_usgs_RDB_service(url, headers, params)
    if verbose:
        print(
            f"Retrieved {params['statReportType']} statistics for site #{site} from {url}"
        )

    header, outputDF, columns, dtype = read_rdb(response.text)

    return hydroRDB(header, outputDF, columns, dtype, response.text)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 26 10:14:38 2016

@author: Marty

test_online_resources.py

This module can be used to check whether various online data servers are up
and returning the expected data.

These tests have all been named so that they will not be included in
the normal test suite.

Nose2 will find these tests, but unittest will not. The continuous integration
and setup.py will use unittests, and therefore won't find this test.
"""
from __future__ import absolute_import, print_function, division, unicode_literals
from hydrofunctions import hydrofunctions as hf


"""
List of Valid URL's for NWIS
paste these into a web browser or use curl to use these URL's.


List of URL's that Return Errors or Specific Conditions

200: good request, but no data have been collected in this period.
https://waterservices.usgs.gov/nwis/iv/?format=json%2C1.1&sites=01541180&parameterCd=00060&period=P10D

400: didn't select a site
 https://waterservices.usgs.gov/nwis/dv/?format=json%2C1.1&parameterCd=00060&startDT=2017-01-01&endDT=2017-12-31

400: site names are too short to be right, should be 8 digits
https://waterservices.usgs.gov/nwis/dv/?format=json%2C1.1&sites=51059%2C51061&parameterCd=00060&startDT=2017-01-01&endDT=2017-12-31

400: wrote countyCd twice
https://waterservices.usgs.gov/nwis/dv/?format=json%2C1.1&countyCd=51059&countyCd=51061&parameterCd=00060&startDT=2017-01-01&endDT=2017-12-31
"""


def will_NWIS_return_200_response():
    """This test should not be run as a part of the normal test suite.
    This is a resource test only!
    Because of the function name, it is not discoverable by Unittest or pytest.
    """

    expected = 200
    response = hf.get_nwis("01585200", "dv", "2001-01-01", "2001-01-02")
    actual = response.status_code
    assert expected == actual
    print("NWIS is up and running!")


if __name__ == "__main__":
    will_NWIS_return_200_response()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_usgs_rdb
----------------------------------

Tests for `usgs_rdb` module.

"""
from __future__ import absolute_import, print_function, division, unicode_literals
from unittest import mock
import unittest

import pandas as pd
import numpy as np
from requests import Response
from requests.exceptions import HTTPError

import hydrofunctions as hf
from .fixtures import (
    fakeResponse,
    field_fixture,
    rating_fixture,
    peaks_fixture,
    parsing_error_fixture,
)


class TestReadRdb(unittest.TestCase):
    """Test the functions that request and parse rdb files from the USGS.

    Test the following topics:
        - ✓Is the correct url formed?
        - What happens when a bad station ID is requested?
            - See Issue #71. a 200 status gets returned, so this error is not caught!
        - ✓does the request header ask for zipped data?
        - Is the # header preserved?
        - Can # header variables be retrieved? (such as found in rating fixture)
        - Are the column names retrieved & parsed?
        - Do the dtypes note date columns? (I think that's the only type reliably noted)
        - Does the dataframe get returned with the proper number of rows and columns?
        - Are the column names recorded properly?
        - Are numbers converted to floats?
        - Are strings recorded as strings?
        - Are dates recorded as dates?
        - Are dates recorded with the proper time zone and converted to UTC?
    """

    def test_read_rdb_returns_4_tuple_of_DF_list_list_str(self):
        test_data = field_fixture
        header, outputDF, columns, dtype = hf.read_rdb(test_data)
        self.assertIs(
            type(outputDF),
            pd.core.frame.DataFrame,
            msg="Read_dbf did not return a dataframe.",
        )
        self.assertIs(
            type(columns), list, msg="read_dbf did not return columns as a list."
        )
        self.assertIs(type(dtype), list, msg="read_dbf did not return dtype as a list.")
        self.assertIs(
            type(header), str, msg="read_dbf did not return header as a string."
        )

    @mock.patch("requests.get")
    def test_get_usgs_RDB_service_returns_response_if_RDB(self, mock_get):
        # Does this call Requests.get properly?
        expected_url = "expected_url"
        expected_status_code = 200
        expected_text = field_fixture

        expected = fakeResponse(code=expected_status_code, text=expected_text)
        mock_get.return_value = expected

        actual = hf.get_usgs_RDB_service(url=expected_url, headers=None, params=None)
        mock_get.assert_called_once_with(expected_url, headers=None, params=None)

    @mock.patch("requests.get")
    def test_get_usgs_RDB_raises_for_200_but_not_RDB(self, mock_get):
        # The function should raise even if status code is 200 but it doesn't return a RDB file.
        expected_url = "expected_url"
        expected_status_code = 200
        expected_text = "not an RDB file"
        expected = fakeResponse(code=expected_status_code, text=expected_text)
        mock_get.return_value = expected

        with self.assertRaises(hf.HydroNoDataError):
            hf.get_usgs_RDB_service(expected_url)

    @mock.patch("requests.get")
    def test_get_usgs_RDB_raises_for_400(self, mock_get):
        # Does this raise for a non-200 status code?
        expected_url = "expected_url"
        failed_response = mock.Mock(Response)
        failed_response.status_code = 400
        failed_response.text = ""
        failed_response.url = expected_url
        failed_response.raise_for_status.side_effect = HTTPError()
        mock_get.return_value = failed_response

        with self.assertRaises(HTTPError):
            hf.get_usgs_RDB_service(expected_url)

    @mock.patch("requests.get")
    def test_site_file(self, mock_get):
        mock_response = mock.Mock(Response)
        mock_response.status_code = 200
        mock_response.text = field_fixture
        mock_response.url = "expected_url"
        mock_get.return_value = mock_response
        actual = hf.site_file("site")
        self.assertIs(
            type(actual.table),
            pd.core.frame.DataFrame,
            msg="site_file did not return a dataframe as expected.",
        )
        self.assertIs(
            type(actual),
            hf.hydroRDB,
            msg="site_file did not return a hydroRDB as expected.",
        )

    @mock.patch("requests.get")
    def test_data_catalog(self, mock_get):
        mock_response = mock.Mock(Response)
        mock_response.status_code = 200
        mock_response.text = field_fixture
        mock_response.url = "expected_url"
        mock_get.return_value = mock_response
        actual = hf.data_catalog("site")
        self.assertIs(
            type(actual.table),
            pd.core.frame.DataFrame,
            msg="data_catalog did not return a dataframe as expected.",
        )
        self.assertIs(
            type(actual),
            hf.hydroRDB,
            msg="data_catalog did not return a hydroRDB as expected.",
        )

    @mock.patch("requests.get")
    def test_field_meas_requests_proper_url(self, mock_get):
        sample_site_id = "666"
        expected_url = (
            "https://waterdata.usgs.gov/nwis/measurements?site_no="
            + sample_site_id
            + "&agency_cd=USGS&format=rdb_expanded"
        )

        expected_status_code = 200
        expected_text = field_fixture

        expected = fakeResponse(code=expected_status_code, text=expected_text)
        expected_headers = {"Accept-encoding": "gzip"}

        expected_params = None

        mock_get.return_value = expected
        actual = hf.field_meas(sample_site_id)
        mock_get.assert_called_once_with(
            expected_url, headers=expected_headers, params=expected_params
        )

    @mock.patch("requests.get")
    def test_field_meas_returns_well_formed_DF(self, mock_get):
        sample_site_id = "026546"

        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"
        expected.text = field_fixture
        expected_rows = 9
        expected_cols = 31  # 31 = 32 cols - 1 col for index.
        expected_col_names = [
            "agency_cd",
            "site_no",
            "measurement_nu",
            "tz_cd",
            "q_meas_used_fg",
            "party_nm",
            "site_visit_coll_agency_cd",
            "gage_height_va",
            "discharge_va",
            "measured_rating_diff",
            "gage_va_change",
            "gage_va_time",
            "control_type_cd",
            "discharge_cd",
            "chan_nu",
            "chan_name",
            "meas_type",
            "streamflow_method",
            "velocity_method",
            "chan_discharge",
            "chan_width",
            "chan_area",
            "chan_velocity",
            "chan_stability",
            "chan_material",
            "chan_evenness",
            "long_vel_desc",
            "horz_vel_desc",
            "vert_vel_desc",
            "chan_loc_cd",
            "chan_loc_dist",
        ]

        expected_row_2 = [
            "USGS",
            "01541200",
            3,
            np.nan,
            "Yes",
            "GAR",
            "USGS",
            6.77,
            3490,
            "Good",
            -0.17,
            1.30,
            "Clear",
            "MEAS",
            1,
            "Imported Channel 1",
            "UNSP",
            "other",
            np.nan,
            3490,
            np.nan,
            np.nan,
            np.nan,
            "UNSP",
            "UNSP",
            "UNSP",
            "unkn",
            "UNSP",
            "UNSP",
            "UNSP",
            np.nan,
        ]

        mock_get.return_value = expected
        actual = hf.field_meas(sample_site_id)

        self.assertIs(
            type(actual.table),
            pd.core.frame.DataFrame,
            msg="field_meas did not return a dataframe as expected.",
        )
        actual_rows, actual_cols = actual.table.shape
        self.assertEqual(
            actual_rows,
            expected_rows,
            msg="field_meas returned the wrong number of rows.",
        )
        self.assertEqual(
            actual_cols,
            expected_cols,
            msg="field_meas returned the wrong number of columns.",
        )
        actual_col_names = actual.table.columns.values.tolist()
        self.assertListEqual(
            actual_col_names,
            expected_col_names,
            msg="field_meas returned the wrong set of column names.",
        )
        actual_row_2 = actual.table.iloc[2].values.tolist()
        # This works better when comparing nans, sometimes.
        np.testing.assert_array_equal(
            actual_row_2,
            expected_row_2,
            err_msg="field_meas returned the wrong values for row 2.",
        )

    # ************
    @mock.patch("requests.get")
    def test_peaks_requests_proper_url(self, mock_get):
        sample_site_id = "666"
        expected_url = (
            "https://nwis.waterdata.usgs.gov/nwis/peak?site_no="
            + sample_site_id
            + "&agency_cd=USGS&format=rdb"
        )

        expected_status_code = 200
        expected_text = peaks_fixture

        expected = fakeResponse(code=expected_status_code, text=expected_text)
        expected_headers = {"Accept-encoding": "gzip"}

        expected_params = None

        mock_get.return_value = expected
        actual = hf.peaks(sample_site_id)
        mock_get.assert_called_once_with(
            expected_url, headers=expected_headers, params=expected_params
        )

    @mock.patch("requests.get")
    def test_peaks_returns_well_formed_DF(self, mock_get):
        sample_site_id = "026546"

        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"
        expected.text = peaks_fixture
        expected_rows = 18
        expected_cols = 12  # 12 = 13 cols - 1 col for index.
        expected_col_names = [
            "agency_cd",
            "site_no",
            "peak_tm",
            "peak_va",
            "peak_cd",
            "gage_ht",
            "gage_ht_cd",
            "year_last_pk",
            "ag_dt",
            "ag_tm",
            "ag_gage_ht",
            "ag_gage_ht_cd",
        ]

        expected_row_2 = [
            "USGS",
            "01542500",
            np.nan,
            "19600",
            np.nan,
            "8.79",
            "2.0",
            np.nan,
            "1941-03-05",
            np.nan,
            "8.95",
            "1.0",
        ]

        mock_get.return_value = expected
        actual = hf.peaks(sample_site_id)

        self.assertIs(
            type(actual.table),
            pd.core.frame.DataFrame,
            msg="field_meas did not return a dataframe as expected.",
        )
        actual_rows, actual_cols = actual.table.shape
        self.assertEqual(
            actual_rows,
            expected_rows,
            msg="field_meas returned the wrong number of rows.",
        )
        self.assertEqual(
            actual_cols,
            expected_cols,
            msg="field_meas returned the wrong number of columns.",
        )
        actual_col_names = actual.table.columns.values.tolist()
        self.assertListEqual(
            actual_col_names,
            expected_col_names,
            msg="field_meas returned the wrong set of column names.",
        )
        actual_row_2 = actual.table.iloc[2].values.tolist()
        # This works better when comparing nans, sometimes.
        np.testing.assert_array_equal(
            actual_row_2,
            expected_row_2,
            err_msg="field_meas returned the wrong values for row 2.",
        )

    # **************

    @mock.patch("requests.get")
    def test_rating_curve_requests_proper_url(self, mock_get):
        sample_site_id = "095695826546"
        expected_url = (
            "https://waterdata.usgs.gov/nwisweb/data/ratings/exsa/USGS."
            + sample_site_id
            + ".exsa.rdb"
        )

        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"
        expected.text = rating_fixture
        expected_headers = {"Accept-encoding": "gzip"}

        expected_params = None

        mock_get.return_value = expected
        actual = hf.rating_curve(sample_site_id)
        mock_get.assert_called_once_with(
            expected_url, headers=expected_headers, params=expected_params
        )

    @mock.patch("requests.get")
    def test_rating_curve_returns_well_formed_DF(self, mock_get):
        sample_site_id = "095695826546"

        expected = fakeResponse()
        expected.status_code = 200
        expected.reason = "any text"
        expected.text = rating_fixture
        expected_rows = 9
        expected_cols = 4
        expected_col_names = ["stage", "shift", "discharge", "stor"]
        expected_row_2 = [2.52, 0.00, 30.65, np.nan]

        mock_get.return_value = expected
        actual = hf.rating_curve(sample_site_id)

        self.assertIs(
            type(actual.table),
            pd.core.frame.DataFrame,
            msg="rating_curve did not return a dataframe as expected.",
        )
        actual_rows, actual_cols = actual.table.shape
        self.assertEqual(
            actual_rows,
            expected_rows,
            msg="rating_curve returned the wrong number of rows.",
        )
        self.assertEqual(
            actual_cols,
            expected_cols,
            msg="rating_curve returned the wrong number of columns.",
        )
        actual_col_names = actual.table.columns.values.tolist()
        self.assertListEqual(
            actual_col_names,
            expected_col_names,
            msg="rating_curve returned the wrong set of column names.",
        )
        actual_row_2 = actual.table.loc[2].values.tolist()
        self.assertListEqual(
            actual_row_2,
            expected_row_2,
            msg="rating_curve returned the wrong values for row 2.",
        )

    @mock.patch("requests.get")
    def test_stats_requests_proper_url(self, mock_get):
        sample_site_id = "095695826546"
        expected_url = "https://waterservices.usgs.gov/nwis/stat/"

        expected = fakeResponse()
        expected.text = rating_fixture
        expected.status_code = 200
        expected_headers = {"Accept-encoding": "gzip"}
        expected_params = {
            "statReportType": "daily",
            "statType": "all",
            "sites": sample_site_id,
            "format": "rdb",
            "parameterCd": "00060",
        }

        mock_get.return_value = expected
        actual = hf.stats(sample_site_id, "daily", parameterCd="00060")
        mock_get.assert_called_once_with(
            expected_url, headers=expected_headers, params=expected_params
        )

    @mock.patch("requests.get")
    def test_peaks_handles_parsing_error(self, mock_get):
        site_id = "06813500"  # This site returns data with a malformed date.
        expected_status_code = 200
        expected_text = parsing_error_fixture
        expected = fakeResponse(code=expected_status_code, text=expected_text)
        mock_get.return_value = expected

        # This will cause a parsing error that must be dealt with.
        actual = hf.peaks(site_id)


class Test_hydroRDB(unittest.TestCase):
    """Test the hydroRDB class.

    - Do the RDB functions (peaks, stats, rating_curve, field_meas) return a hydroRDB object?
    - Does the hydroRDB have a proper repr?
    - Does the hydroRDB have a working __iter__ function?
    """

    def test_hydroRDB_is_obj(self):
        header = "expected header"
        table = "expected table"
        columns = "expected columns"
        dtypes = "expected dtypes"
        rdb = "expected rdb file"
        actual = hf.hydroRDB(header, table, columns, dtypes, rdb)
        self.assertIsInstance(actual, hf.hydroRDB)

    def test_hydroRDB_has_properties_and_methods(self):
        header = "expected header"
        table = "expected table"
        columns = "expected columns"
        dtypes = "expected dtypes"
        rdb = "expected rdb file"
        actual = hf.hydroRDB(header, table, columns, dtypes, rdb)
        self.assertIsInstance(actual, hf.hydroRDB)
        self.assertEqual(actual.header, header)
        self.assertEqual(actual.table, table)
        self.assertEqual(actual.columns, columns)
        self.assertEqual(actual.dtypes, dtypes)
        self.assertEqual(actual.rdb, rdb)
        actual_repr = actual.__repr__()
        # actual._repr_html_() requires that table is a dataframe with a _repr_html_()
        # actual_html_repr = actual._repr_html_()

    def test_hydroRDB_can_return_tuple(self):
        header = "expected header"
        table = pd.DataFrame({"A": [1, 2], "B": [3, 4]})
        columns = "expected columns"
        dtypes = "expected dtypes"
        rdb = "expected rdb file"
        actual = hf.hydroRDB(header, table, columns, dtypes, rdb)
        self.assertIsInstance(actual, hf.hydroRDB)
        head, table = actual
        self.assertIsInstance(head, str)
        self.assertIsInstance(table, pd.DataFrame)



"""
hydrofunctions.charts
~~~~~~~~~~~~~~~~~~~~~

This module contains charting functions for Hydrofunctions.

-----
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import matplotlib

# Recommended that I use this line to avoid errors in TravisCI
# See https://matplotlib.org/faq/howto_faq.html
# Basically, matplotlib usually uses an 'X11 connection' by default; Travis CI
# does not have this configured, so you need to set your backend explicitly.
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter
import numpy as np
import pandas as pd


def flow_duration(
    Qdf,
    xscale="logit",
    yscale="log",
    ylabel="Stream Discharge (m³/s)",
    symbol=".",
    legend=True,
    legend_loc="best",
    title="",
):
    """Creates a flow duration chart from a dataframe of discharges.

    Args:
        Qdf (dataframe): a dataframe of discharge values.

            * Values should be arranged in columns
            * No sorting necessary
            * Rows do not need an index
            * If more than one column, each column will be added as a\
                separate color to the chart.
            * Only include columns with discharge values; no metadata

        xscale (str, 'logit' | 'linear'): Type of x scale for plotting probabilities
            default is 'logit', so that each standard deviation is nearly the
            same distance on the x scale. 'linear' is the other option.

        yscale (str, 'log' | 'linear'): The type of y scale for plotting discharge.
            Default is 'log'.

        ylabel (`str`, default `'Stream Discharge (ft³/s)'`): The label for the Y axis.

        xlabel (not implemented)

        symbol (str, '.' | ','): formatting symbol for points.

            * point: '.' (default)
            * pixel point: ','
            * circle: 'o'
            * triangle up: '^'

            See https://matplotlib.org/api/markers_api.html for full list of
            point formatters.

        legend (`bool`, default `True`): Whether the legend should be plotted.

        legend_loc (`str`, default `best`): the location of the legend.

            * 'best': Automatically choose the option below with the least overlap.
            * 'upper left', 'upper right', 'lower left', 'lower right': place the legend at the corresponding corner of the axes/figure.
            * 'upper center', 'lower center', 'center left', 'center right': place the legend at the center of the corresponding edge of the axes/figure.
            * 'center': place the legend at the center of the axes/figure.
            * The location can also be a 2-tuple giving the coordinates of the lower-left corner of the legend in axes coordinates.

        title (`str`, default ''): Text to use as a figure title. If no text
            is provided, no title will be created (default).

    Returns:
        fig, ax (`matplotlib.figure.Figure`, `matplotlib.axes.Axes`):
            Returns a tuple that includes a matplotlib 'figure' and 'axes'. The figure
            is a container with all of the drawing inside of it; the axes are an array
            of matplotlib charts. Together, they will plot immediately in a Jupyter
            notebook if the command `%matplotlib inline` was previously issued.
            The figure and axes may be altered after they are returned.

    """
    rank = Qdf.rank(ascending=False, pct=True)
    x = rank
    y = Qdf
    fig, ax = plt.subplots(1, 1)
    for column in Qdf.columns.values:
        ax.plot(x.loc[:, column], y.loc[:, column], symbol, label=column)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_ylabel(ylabel)
    if legend:
        ax.legend(loc=legend_loc)
    if title:
        ax.title.set_text(title)
    ax.set_xlabel("Probability of Exceedence")
    ax.xaxis.set_minor_formatter(NullFormatter())
    return fig, ax


def cycleplot(
    Qseries,
    cycle="diurnal",
    compare=None,
    y_label="Discharge (ft³/s)",
    legend=True,
    legend_loc="best",
    title="",
):
    """Creates a chart to illustrate annual and diurnal cycles.

    This chart will use the pandas groupby method to plot the mean and median
    values for a time-indexed dataframe. It helps you identify diurnal patterns
    by plotting the mean and median values over 24 hours for a diurnal pattern,
    and over a year for annual patterns.

    This function will also use the 'compare' argument to create a series of
    charts to compare how well these cycles appear in different groups. For
    example, is the diurnal cycle more visible in December versus June? In this
    case, you would use::

        hf.cycleplot(myDataFrame, cycle='diurnal', compare = 'month')

    This will produce twelve charts, each covering 24 hours. A line will
    represent the mean values over 24 hours, another line represents the
    median, and two grey stripes represent the 0.4 to 0.6 quantile, and the
    0.2 to 0.8 quantile range.

    Args:
        Qseries (series): a Pandas series of discharge values.

            * Values should be arranged in columns
            * Should use a dateTimeIndex

        cycle (str): The period of the cycle to be illustrated, along with the
            method for binning. The options are:

            * diurnal (default): plots the values for a 24 hour cycle.
            * diurnal-smallest: uses the smallest increment of time \
            available to bin the time units for a 24 hour cycle.
            * diurnal-hour: uses hours to bin measurements for a 24-hour \
            cycle.
            * annual: plots values into a 365 day cycle.
            * annual-day: the annual cycle using 365 day-long bins.
            * annual-week: the annual cycle using 52 week-long bins.
            * annual-month: the annual cycle using 12 month-long bins.
            * weekly: a 7-day cycle using seven 24-hour long bins. Note \
            that unlike the others, this is not a natural cycle, and is \
            likely has anthropogenic origins.

        compare (str): The system for splitting the data into
            groups for a set of comparison charts.

            * None (default): No comparison will be made; only one chart.
            * month: twelve plots will be produced, one for each month.
            * weekday: seven plots will be produced, one for each day of \
            the week.
            * weekend: two plots will be produced, one for the five weekdays, \
            one for Saturday and Sunday.
            * night: two plots will be produced, one for night (6pm to 6am), \
            one for day (6am to 6pm).

        y_label (str): The label for the y axis.

        legend (bool): default True. Whether the legend should be plotted.

        legend_loc (str): default is 'best'. The location of the legend.

            * 'best': Automatically choose the option below with the least overlap.
            * 'upper left', 'upper right', 'lower left', 'lower right': place the legend at the corresponding corner of the axes/figure.
            * 'upper center', 'lower center', 'center left', 'center right': place the legend at the center of the corresponding edge of the axes/figure.
            * 'center': place the legend at the center of the axes/figure.
            * The location can also be a 2-tuple giving the coordinates of the lower-left corner of the legend in axes coordinates.

        title (str): default is ''. Text to use as a figure title. If no text
            is provided, no title will be created (default).

    Returns:
        fig, ax (matplotlib.figure.Figure, matplotlib.axes.Axes):
            Returns a tuple that includes a matplotlib 'figure' and 'axes'. The figure
            is a container with all of the drawing inside of it; the axes are an array
            of matplotlib charts. Together, they will plot immediately in a Jupyter
            notebook if the command `%matplotlib inline` was previously issued.
            The figure and axes may be altered after they are returned.

    Note:
        inspired by https://jakevdp.github.io/PythonDataScienceHandbook/03.11-working-with-time-series.html
        Jake VanderPlas. 2016. Python Data Science Handbook. O'Reilly Media, Inc.
    """
    # is Qseries a series of discharge? if not, raise error.
    # or, select the discharge...
    if isinstance(Qseries, pd.DataFrame):
        # Qseries = Qseries.iloc[:,0] # Select first column; but this is not always numeric.
        # Select the first of the numeric columns.
        Qseries = Qseries.loc[:, Qseries.select_dtypes(include="number").columns[0]]
    if not isinstance(Qseries, pd.Series):
        raise ValueError(
            "Cycleplot only accepts a single series of data as "
            f" an argument. You supplied a {type(Qseries)}."
        )

    if cycle == "annual":
        # aggregate into 365 bins to show annual cycles. Same as annual-date
        cycleby = Qseries.index.dayofyear
        x_label = " (day # of the year)"
    elif cycle == "annual-date":
        cycleby = Qseries.index.dayofyear
        x_label = " (day # of the year)"
    elif cycle == "annual-week":
        # aggregate into 52 bins to show annual cycles.
        cycleby = Qseries.index.week
        x_label = " (week # of the year)"
    elif cycle == "annual-month":
        # aggregate into 12 binds to show annual cycles.
        cycleby = Qseries.index.month
        x_label = " (month # of the year)"
    elif cycle == "weekly":
        # aggregate into 7 bins to show week-long cycles.
        # Note: 7-day cycles are not natural cycles.
        cycleby = Qseries.index.weekday
        x_label = " (day of the week, Monday = 0)"
    elif cycle == "diurnal":
        # aggregate into 24 bins to show 24-hour daily (diurnal) cycles.
        cycleby = Qseries.index.hour
        x_label = " (hour of the day)"
    #    elif cycle == 'diurnal-smallest':
    # Uses the smallest unit available in the time index to show 24-hour diurnal cycles.
    #        cycleby = Qseries.index.time
    #        x_label = ' (time of day)'
    elif cycle == "diurnal-hour":
        # aggregate into 24 bins to show 24-hour daily (diurnal) cycles.
        cycleby = Qseries.index.hour
        x_label = " (hour of the day)"
    else:
        raise ValueError(
            "The cycle label '", cycle, "' is not recognized as an option."
        )

    if compare is None:
        # Don't make a comparison plot.
        # TODO: This is a silly categorization to force all values in the index into the same category.
        compareby = np.where(
            Qseries.index.weekday < 20, "A", "B"
        )  # Since no comparison is desired, this puts all of the data into group A.
        sub_titles = [""]
    elif compare == "month":
        # Break the time series into 12 months to compare to each other.
        compareby = Qseries.index.month
        sub_titles = [
            "January",
            "February",
            "March",
            "April",
            "May",
            "June",
            "July",
            "August",
            "September",
            "October",
            "November",
            "December",
        ]
    elif compare == "weekday":
        # Break the time series into 7 days of the week to compare to each other.
        compareby = Qseries.index.weekday
        sub_titles = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
    elif compare == "weekend":
        # Break the time series into 2 groups, Weekdays and Weekends for comparison.
        compareby = np.where(Qseries.index.weekday < 5, "Weekday", "Weekend")
        sub_titles = ["Weekdays", "Weekends"]
    elif compare == "night":
        # Break the time series into 2 groups to compare day versus night.
        # TODO: This assumes the index hour is in local time, but it is in UTC time
        compareby = np.where(
            (Qseries.index.hour >= 6) & (Qseries.index.hour < 19), "Day", "Night"
        )
        sub_titles = ["Day", "Night"]
    else:
        print(
            "The compare label '",
            compare,
            "' is not recognized as an option. Using compare=None instead.",
        )
        compareby = np.where(
            Qseries.index.weekday < 20, "A", "B"
        )  # Since no comparison is desired, this puts all of the data into group A.
        sub_titles = ["data"]

    selection = [compareby, cycleby]
    compare_list = list(np.unique(compareby))

    # group first by the compareby series, then by the cycle.
    grouped = Qseries.groupby(selection)

    # complicated_multiindex_DF_of_results = Qseries.groupby(selection).agg('mean', q2, q4, 'median', q6, q8)
    #    results2 = Qseries.groupby(selection).agg(
    #            mean=
    #            Q2=
    #            Q4=
    #            Q5=
    #            Q6=
    #            Q8=
    #            )

    #    # Method 3 =================================
    #
    #    Q2 = np.percentile(grouped, .2)
    #    Q4 = np.percentile(grouped, .4)
    #    Q5 = np.percentile(grouped, .5)
    #    Q6 = np.percentile(grouped, .6)
    #    Q8 = np.percentile(grouped, .8)
    #
    #    # Method 2====================================================
    #    # Why is this necessary? Pandas 0.25.0 won't let you do this:
    #    # grouped.quantile()  anymore. !?!
    #    def q2(x):
    #        return x.quantile(.2)
    #    def q4(x):
    #        return x.quantile(.4)
    #    def q6(x):
    #        return x.quantile(.6)
    #    def q8(x):
    #        return x.quantile(.8)
    #
    #    mean = grouped.mean()
    #    Q2 = grouped.agg(q2)  #    Q2 = DF.groupby(selection).quantile(.2)
    #    Q4 = grouped.agg(q4)  #    Q4 = grouped.quantile(0.4)
    #    Q5 = grouped.median() #    Q5 = grouped.quantile(.5)
    #    Q6 = grouped.agg(q6)  #    Q6 = grouped.quantile(.6)
    #    Q8 = grouped.agg(q8)  #    Q8 = grouped.quantile(.8)
    #
    # Method 1 =================================
    mean = grouped.mean()
    Q2 = grouped.quantile(0.2)
    Q4 = grouped.quantile(0.4)
    Q5 = grouped.quantile(0.5)
    Q6 = grouped.quantile(0.6)
    Q8 = grouped.quantile(0.8)
    # ==============================

    Nplots = len(compare_list)
    fig, axs = plt.subplots(1, Nplots, figsize=(14, 6), sharey=True, sharex=True)
    if Nplots == 1:
        # If there is only one subplot, it gets returned as a single subplot instead of as a numpy array. In this case, we convert it to an array.
        axs = np.array([axs])

    for i, item in enumerate(compare_list):
        axs[i].plot(mean.loc[item], label="mean")
        # axs[i].plot(Q2.loc[item], label='20th percentile', color='black', linestyle='dotted', linewidth=2)
        axs[i].plot(
            Q5.loc[item], label="median", color="black", linestyle="dotted", linewidth=2
        )
        # axs[i].plot(Q8.loc[item], label='80th percentile', color='grey', linestyle='dashed', linewidth=1)
        axs[i].fill_between(
            Q2.loc[item].index,
            Q2.loc[item].values.flatten(),
            Q8.loc[item].values.flatten(),
            facecolor="grey",
            alpha=0.5,
        )
        axs[i].fill_between(
            Q4.loc[item].index,
            Q4.loc[item].values.flatten(),
            Q6.loc[item].values.flatten(),
            facecolor="grey",
            alpha=0.5,
        )
        axs[i].set_title(sub_titles[i])
        # axs[i].xaxis.set_major_locator(plt.MaxNLocator(4))
        # axs[i].xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%H'))

    # Set the legend on either the ax or fig.
    if legend:
        axs[0].legend(loc=legend_loc, fancybox=True, framealpha=0.5)
    # fig.legend(loc='upper center', shadow=True, frameon=True, fancybox=True, framealpha=0.5)

    # Get the yaxis limits, set bottom to zero.
    ymin, ymax = axs[0].get_ylim()
    axs[0].set_ylim(0, ymax)
    axs[0].set_ylabel(y_label)
    axs[0].set_xlabel("Time" + x_label)
    plt.tight_layout()
    if title:
        fig.suptitle(title)

    return fig, axs



"""
hydrofunctions.logging
~~~~~~~~~~~~~~~~~~~~~~

This module contains the tools used for internal diagnostic logging.

Logging is disabled by default. Users can start logging by using the 
`hf._start_logging()`  function. This will create a file "hydrofunctions_testing.log" in
the main directory. This function also allows users to set the level of severity that
will be logged. The default is to capture all messages, including the lowest level
'DEBUG' messages.

To create log messages within a module, follow these steps:

1. Create a custom logger for the module.
    - Place the statement `logger = logging.getLogger(__name__)` at the top of the module.
    - This will create a custom logger that is named after the module.
    - call the logger like this: logger.info("Hello!")

2. Log a message within your code.
    - Create a message: `msg = "This is the text of the message."`
    - Include the value of important variables
    - There is no need to include the time or name of the module or function. These are included in the standard message format.
    - Decide on a 'level' for the message:
        - DEBUG: this is the lowest level; for tracking ordinary internal values
        - INFO: internal or user events that are working as expected
        - WARNING: situations where back-up procedures are needed, unexpected\
            situations and possibly ordinary exceptions that have been caught
        - ERROR: internal problems that prevent the software from completing an action
        - CRITICAL: serious errors that cause the shutdown of the software
    - Add the message to the log file: `logger.info(msg)`
    - Each level of severity has its own method: .debug(), .info(), .critical(), etc.
    - HydroExceptions such as HydroNoDataError generate their own error logs.
    - It might be useful to log other errors when they are raised.

3. Start the logging system.
    - Logging is off by default.
    - To start logging, call hydrofunctions._start_logging()
    - You can specify the level that will be captured in the log with the `loglevel` parameter.
    - Set level like this: `hf._start_logging('info')`  (Case does not matter)
    - The default is to capture from the lowest level (DEBUG) up
    - Starting the logging system will create a new file "hydrofunctions_testing.log" if\
        it doesn't already exist; if it does, it will add new messages at the bottom\
        under a start up message to the 'root' module.

4. Read the log.
    - The file, "hydrofunctions_testing.log" will appear in the root directory
    - All messages from hydrofunctions will have the following:
        - timestamp
        - name of logger: 'root' for the start message, all others should be named for\
            the module that creates the message
        - levelname: the level of the message ('DEBUG', 'INFO', etc)
        - funcName: the name of the function that sent the message to the log
        - message: the message generated by the logger function: `logger.info("Hello!")`
    - The first message created by the `hf._start_logging()` will be from 'root'
    - Messages from dependencies will be captured too.

-----
"""
import logging


def _start_logging(loglevel="DEBUG"):
    """Create a log file if it doesn't exist and start logging messages.

    Args:
        loglevel (str):
            The level of message that should be captured in the log. Valid values
            are (from lowest to highest):
            - 'DEBUG': (default) Detailed information, typically of interest only when diagnosing problems.
            - 'INFO': Confirmation that things are working as expected.
            - 'WARNING': An indication that something unexpected happened, or indicative of some problem in the near future (e.g. ‘disk space low’). The software is still working as expected.
            - 'ERROR': Due to a more serious problem, the software has not been able to perform some function.
            - 'CRITICAL': A serious error, indicating that the program itself may be unable to continue running.

    Raises:
        ValueError: if loglevel does not correspond to one of the five valid levels
    """
    # This code is from Vinay Sajip in the Python.org logging tutorial.
    # https://docs.python.org/3/howto/logging.html
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % loglevel)
    format = "%(asctime)s — %(name)s — %(levelname)s — %(funcName)s — %(message)s"
    logging.basicConfig(
        filename="hydrofunctions.log",
        level=numeric_level,
        format=format,
    )
    logging.captureWarnings(True)
    logging.info(
        f"Logging started. Messages at the {loglevel} level and above will be captured.\nLine format is: {format}"
    )



"""
test_logging
------------

Tests for the `logging` module.
"""
import logging
from unittest import mock
import unittest
from unittest import mock
import warnings

import hydrofunctions as hf

hf._start_logging()


class TestLogging(unittest.TestCase):
    """Test the hydrofunctions logging system.

    test the following:
        Is logging off by default?
        There should be no files generated if logging is off
        The file "hydrofunctions.log" should be generated when logging is started
        ✔HydroExceptions should generate a log message
        ✔loggers should be able to post a message to the log
        ✔hf._start_logging() calls logging.basicConfig() properly
        ✔warning messages should be logged.

    """

    def test_logging_message(self):
        logger = logging.getLogger(__name__)
        expected_message = "test_logging"

        with self.assertLogs() as captured:
            logger.info(expected_message)
            expected_level = logging.INFO

        actual_n_of_messages = len(captured.records)
        actual_message = captured.records[0].getMessage()
        actual_level = captured.records[0].levelno
        # The actual_log_line is hard to test against until you set the output format.
        # actual_log_line = captured.output
        # print(f"***********/n/n/ncaptured log line: /n{actual_log_line}/n/n/n***********/n")
        self.assertEqual(actual_n_of_messages, 1, "There should only be one message")
        self.assertEqual(actual_message, expected_message, "Wrong message was logged")
        self.assertEqual(actual_level, expected_level, "Message logged at wrong level")

    def test_logging_HydroExceptions_log_message(self):
        expected_message = "test_logging_HydroExceptions_log_msg"
        with self.assertLogs() as captured:
            try:
                raise hf.HydroNoDataError(expected_message)
            except hf.HydroNoDataError:
                pass

        expected_level = logging.ERROR
        actual_level = captured.records[0].levelno
        self.assertEqual(actual_level, expected_level, "Message logged at wrong level")
        actual_message = captured.records[0].getMessage()
        self.assertEqual(actual_message, expected_message, "Wrong message was logged")

    @mock.patch("logging.basicConfig")
    def test_logging_start_logging_logs_warnings(self, mock_config):
        # logger = logging.getLogger(__name__)
        expected_message = "Test that warnings are logged."
        # hf._start_logging()
        # mock_config.assert_called()

        with self.assertLogs() as captured:
            warnings.warn(expected_message, hf.HydroUserWarning)
        actual_message = captured.records[0].getMessage()
        # warning logs have their own formatting that contains the actual_message.
        self.assertIn(
            expected_message,
            actual_message,
            f"Wrong message was logged. Actual: {actual_message}",
        )

    @mock.patch("logging.basicConfig")
    def test_logging_start_logging_calls_basicConfig(self, mock_config):
        expected_level = logging.ERROR
        hf._start_logging("error")
        expected_filename = "hydrofunctions.log"

        # This is a bit brittle: the format is not set in stone.
        # expected format = "%(asctime)s — %(name)s — %(levelname)s — %(funcName)s — %(message)s"
        expected_format = mock.ANY
        # .assert_called_once_with() passes in pytest, but fails in unittest- called 3 times.
        # pytest prevents dependencies from calling logging.basicConfig(); unittest doesn't
        mock_config.assert_any_call(
            filename=expected_filename, level=expected_level, format=expected_format
        )



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 15 10:24:48 2019

@author: Marty
"""
# hf.NWIS('01541200')
# https://waterservices.usgs.gov/nwis/dv/?format=json%2C1.1&sites=01541200

# This is a request that doesn't specify a starting date or a period, so NWIS
# returns the most recent readings.
# Amuseingly, for this site, that means you get a temperature reading from
# 1961 as well as discharge from today.
recent_only = {
    "name": "ns1:timeSeriesResponseType",
    "declaredType": "org.cuahsi.waterml.TimeSeriesResponseType",
    "scope": "javax.xml.bind.JAXBElement$GlobalScope",
    "value": {
        "queryInfo": {
            "queryURL": "http://waterservices.usgs.gov/nwis/dv/format=json%2C1.1&sites=01541200",
            "criteria": {
                "locationParam": "[ALL:01541200]",
                "variableParam": "ALL",
                "parameter": [],
            },
            "note": [
                {"value": "[ALL:01541200]", "title": "filter:sites"},
                {
                    "value": "[mode=LATEST, modifiedSince=null]",
                    "title": "filter:timeRange",
                },
                {"value": "methodIds=[ALL]", "title": "filter:methodId"},
                {"value": "2019-04-15T14:23:12.311Z", "title": "requestDT"},
                {"value": "00165560-5f8a-11e9-9824-6cae8b6642f6", "title": "requestId"},
                {
                    "value": "Provisional data are subject to revision. Go to http://waterdata.usgs.gov/nwis/help/?provisional for more information.",
                    "title": "disclaimer",
                },
                {"value": "caas01", "title": "server"},
            ],
        },
        "timeSeries": [
            {
                "sourceInfo": {
                    "siteName": "WB Susquehanna River near Curwensville, PA",
                    "siteCode": [
                        {"value": "01541200", "network": "NWIS", "agencyCode": "USGS"}
                    ],
                    "timeZoneInfo": {
                        "defaultTimeZone": {
                            "zoneOffset": "-05:00",
                            "zoneAbbreviation": "EST",
                        },
                        "daylightSavingsTimeZone": {
                            "zoneOffset": "-04:00",
                            "zoneAbbreviation": "EDT",
                        },
                        "siteUsesDaylightSavingsTime": True,
                    },
                    "geoLocation": {
                        "geogLocation": {
                            "srs": "EPSG:4326",
                            "latitude": 40.9614471,
                            "longitude": -78.5191906,
                        },
                        "localSiteXY": [],
                    },
                    "note": [],
                    "siteType": [],
                    "siteProperty": [
                        {"value": "ST", "name": "siteTypeCd"},
                        {"value": "02050201", "name": "hucCd"},
                        {"value": "42", "name": "stateCd"},
                        {"value": "42033", "name": "countyCd"},
                    ],
                },
                "variable": {
                    "variableCode": [
                        {
                            "value": "00010",
                            "network": "NWIS",
                            "vocabulary": "NWIS:UnitValues",
                            "variableID": 45807042,
                            "default": True,
                        }
                    ],
                    "variableName": "Temperature, water, &#176;C",
                    "variableDescription": "Temperature, water, degrees Celsius",
                    "valueType": "Derived Value",
                    "unit": {"unitCode": "deg C"},
                    "options": {
                        "option": [
                            {
                                "value": "Random Instantaneous Values",
                                "name": "Statistic",
                                "optionCode": "00011",
                            }
                        ]
                    },
                    "note": [],
                    "noDataValue": -999999.0,
                    "variableProperty": [],
                    "oid": "45807042",
                },
                "values": [
                    {
                        "value": [
                            {
                                "value": "11.7",
                                "qualifiers": ["A"],
                                "dateTime": "1961-09-29T00:00:00.000",
                            }
                        ],
                        "qualifier": [
                            {
                                "qualifierCode": "A",
                                "qualifierDescription": "Approved for publication -- Processing and review completed.",
                                "qualifierID": 0,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            }
                        ],
                        "qualityControlLevel": [],
                        "method": [{"methodDescription": "", "methodID": 118850}],
                        "source": [],
                        "offset": [],
                        "sample": [],
                        "censorCode": [],
                    }
                ],
                "name": "USGS:01541200:00010:00011",
            },
            {
                "sourceInfo": {
                    "siteName": "WB Susquehanna River near Curwensville, PA",
                    "siteCode": [
                        {"value": "01541200", "network": "NWIS", "agencyCode": "USGS"}
                    ],
                    "timeZoneInfo": {
                        "defaultTimeZone": {
                            "zoneOffset": "-05:00",
                            "zoneAbbreviation": "EST",
                        },
                        "daylightSavingsTimeZone": {
                            "zoneOffset": "-04:00",
                            "zoneAbbreviation": "EDT",
                        },
                        "siteUsesDaylightSavingsTime": True,
                    },
                    "geoLocation": {
                        "geogLocation": {
                            "srs": "EPSG:4326",
                            "latitude": 40.9614471,
                            "longitude": -78.5191906,
                        },
                        "localSiteXY": [],
                    },
                    "note": [],
                    "siteType": [],
                    "siteProperty": [
                        {"value": "ST", "name": "siteTypeCd"},
                        {"value": "02050201", "name": "hucCd"},
                        {"value": "42", "name": "stateCd"},
                        {"value": "42033", "name": "countyCd"},
                    ],
                },
                "variable": {
                    "variableCode": [
                        {
                            "value": "00060",
                            "network": "NWIS",
                            "vocabulary": "NWIS:UnitValues",
                            "variableID": 45807197,
                            "default": True,
                        }
                    ],
                    "variableName": "Streamflow, ft&#179;/s",
                    "variableDescription": "Discharge, cubic feet per second",
                    "valueType": "Derived Value",
                    "unit": {"unitCode": "ft3/s"},
                    "options": {
                        "option": [
                            {
                                "value": "Mean",
                                "name": "Statistic",
                                "optionCode": "00003",
                            }
                        ]
                    },
                    "note": [],
                    "noDataValue": -999999.0,
                    "variableProperty": [],
                    "oid": "45807197",
                },
                "values": [
                    {
                        "value": [
                            {
                                "value": "717",
                                "qualifiers": ["P"],
                                "dateTime": "2019-04-14T00:00:00.000",
                            }
                        ],
                        "qualifier": [
                            {
                                "qualifierCode": "P",
                                "qualifierDescription": "Provisional data subject to revision.",
                                "qualifierID": 0,
                                "network": "NWIS",
                                "vocabulary": "uv_rmk_cd",
                            }
                        ],
                        "qualityControlLevel": [],
                        "method": [{"methodDescription": "", "methodID": 118849}],
                        "source": [],
                        "offset": [],
                        "sample": [],
                        "censorCode": [],
                    }
                ],
                "name": "USGS:01541200:00060:00003",
            },
        ],
    },
    "nil": False,
    "globalScope": True,
    "typeSubstituted": False,
}



"""
hydrofunctions.station
~~~~~~~~~~~~~~~~~~~~~~

This module contains the Station and NWIS classes, which are used for
organizing and managing data for data collection sites.

-----
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import re

import json
import warnings
from . import hydrofunctions as hf

import numpy as np


class Station(object):
    """A class for organizing stream gauge data for a single request."""

    station_dict = {}

    def __init__(self, site=None):
        Station.station_dict[site] = self
        self.site = site
        # One option is to make it so that you can pass in a get_data function
        # during the creation of an instance.


class NWIS(Station):
    """A class for working with data from the USGS NWIS service.

    Args:
        site (str or list of strings):
            a valid site is '01585200' or ['01585200', '01646502']. Default is
            `None`. If site is not specified, you will need to select sites using
            stateCd or countyCd.

        service (str):
            can either be 'iv' or 'dv' for instantaneous or daily data.
                * 'dv'(default): daily values. Mean value for an entire day.
                * 'iv': instantaneous value measured at this time. Also known\
                as 'Real-time data'. Can be measured as often as every\
                five minutes by the USGS. 15 minutes is more typical.

        start_date (str):
           should take on the form 'yyyy-mm-dd'

        end_date (str):
            should take on the form 'yyyy-mm-dd'

        stateCd (str):
            a valid two-letter state postal abbreviation, such as 'MD'. Default
            is None. Selects all stations in this state. Because this type of
            site selection returns a large number of sites, you should limit
            the amount of data requested for each site.

        countyCd (str or list of strings):
            a valid county FIPS code. Default is None. Requests all stations
            within the county or list of counties. See https://en.wikipedia.org/wiki/FIPS_county_code
            for an explanation of FIPS codes.

        bBox (str, list, or tuple):
            a set of coordinates that defines a bounding box.
                * Coordinates are in decimal degrees.
                * Longitude values are negative (west of the prime meridian).
                * Latitude values are positive (north of the equator).
                * comma-delimited, no spaces, if provided as a string.
                * The order of the boundaries should be: "West,South,East,North"
                * Example: "-83.000000,36.500000,-81.000000,38.500000"

        parameterCd (str or list of strings):
            NWIS parameter code. Usually a five digit code. Default is 'all'.
            A valid code can also be given as a list: parameterCd=['00060','00065']
            This will request data for this parameter.

                * if value is 'all', or no value is submitted, then NWIS will \
                return every parameter collected at this site. (default option)
                * stage: '00065'
                * discharge: '00060'
                * not all sites collect all parameters!
                * See https://nwis.waterdata.usgs.gov/usa/nwis/pmcodes for full list

        period (str):
            NWIS period code. Default is None.
                * Format is "PxxD", where xx is the number of days before \
                today, with a maximum of 999 days accepted.
                * Either use start_date or period, but not both.

        file (str):
            A filename for acting as a cache for the data request. Accepts file
            extensions of '*.json.gz' (default) and '*.parquet'. If this parameter is
            included, the NWIS object will first attempt to read its data from the file.
            If the file does not exist, it will use the other parameters to obtain the
            data and will then save to the provided filename.

            Zipped JSON files will save the original WaterML JSON provided by the NWIS.
            Parquet files will save the dataframe and the metadata for the NWIS object.

        verbose (bool):
            Print output for actions such as making data requests. Default is True.
    """

    def __init__(
        self,
        site=None,
        service="dv",
        start_date=None,
        end_date=None,
        stateCd=None,
        countyCd=None,
        bBox=None,
        parameterCd="all",
        period=None,
        file=None,
        verbose=True,
    ):

        self.ok = False
        if file:
            if len(file.split(".")) == 1:
                file = file + ".json.gz"
            try:
                self.read(file)
                self.ok = True
                if verbose:
                    print("Reading data from", file)

            except OSError as err:
                # File does not exist yet, we'll make it later.
                pass

        if not self.ok:
            self.response = hf.get_nwis(
                site,
                service,
                start_date,
                end_date,
                stateCd=stateCd,
                countyCd=countyCd,
                bBox=bBox,
                parameterCd=parameterCd,
                period=period,
                verbose=verbose,
            )
            try:
                self.json = self.response.json()
                self._dataframe, self.meta = hf.extract_nwis_df(self.json)
                self.ok = self.response.ok
                if file is not None:
                    self.save(file)
                    if verbose:
                        print("Saving data to", file)
            except json.JSONDecodeError as err:
                self.ok = False
                print(f"JSON decoding error. URL: {self.response.url}")
                raise err

        # Can I get rid of this, and only keep metadata in the meta dict?
        if self.ok:
            self.site = site
            self.service = service
            self.start_date = start_date
            self.end_date = end_date
            self.start = self._dataframe.index.min()
            self.end = self._dataframe.index.max()

    def __repr__(self):
        repr_string = ""
        for site_id in sorted(self.meta.keys()):
            repr_string += site_id + ": " + self.meta[site_id]["siteName"] + "\n"
            for param in sorted(self.meta[site_id]["timeSeries"].keys()):
                repr_string += (
                    "    "
                    + param
                    + ": "
                    + self.meta[site_id]["timeSeries"][param]["variableFreq"]
                    + "  "
                    + self.meta[site_id]["timeSeries"][param]["variableDescription"]
                    + " "
                    + self.meta[site_id]["timeSeries"][param]["methodDescription"]
                    + "\n"
                )
        repr_string += "Start: " + str(self.start) + "\n" + "End:   " + str(self.end)
        return repr_string

    def df(self, *args):
        """
        Return a subset of columns from the dataframe.

        Args:
            '': If no args are provided, the entire dataframe will be returned.

            str 'all': the entire dataframe will be returned.

            str 'data': all of the parameters will be returned, with no flags.

            str 'flags': Only the _qualifier flags will be returned. Unless the \
            flags arg is provided, only data columns will be returned. Visit \
            https://waterdata.usgs.gov/usa/nwis/uv?codes_help#dv_cd1 to see a \
            more complete listing of possible codes.

            str 'discharge' or 'q': discharge columns ('00060') will be returned.

            str 'stage': Gauge height columns ('00065') will be returned.

            str any five digit number: any matching parameter columns will be returned. '00065' returns stage, for example.

            str any eight to twelve digit number: any matching stations will be returned.
        """
        all_cols = self._dataframe.columns != ""  # all true
        no_cols = ~all_cols  # all false
        data_cols = self._dataframe.columns.str.contains(
            r"[0-9]$"
        )  # Data columns end in a number.
        flag_cols = self._dataframe.columns.str.contains("_qualifiers")
        Q_cols = self._dataframe.columns.str.contains(
            ":00060:"
        )  # This includes data & flags
        stage_cols = self._dataframe.columns.str.contains(":00065:")
        param_re = r"^\d{5}$"  # parameters are a five-digit number.
        station_re = r"\d{8,12}$"  # station ID's are between 8 and 12 digits.

        sites = no_cols
        params = no_cols
        meta = no_cols
        if len(args) == 0:  # If no args are given, return every column.
            sites = all_cols
            params = all_cols
            meta = all_cols
        else:
            for item in args:
                if item == "all":
                    sites = all_cols
                    params = all_cols
                    meta = all_cols
                    break  # If one param is 'all', ignore the other params and deliver everything.
                elif item == "discharge":
                    if not np.any(['00060' in x for x in list(self._dataframe.columns)]):
                        raise ValueError(
                            "The parameter '{param}' is not contained in this dataset.".format(
                                param=item
                            )
                        )
                    params = Q_cols | params
                elif item == "q":
                    if not np.any(['00060' in x for x in list(self._dataframe.columns)]):
                        raise ValueError(
                            "The parameter '{param}' is not contained in this dataset.".format(
                                param=item
                            )
                        )
                    params = Q_cols | params
                elif item == "stage":
                    if not np.any(['00065' in x for x in list(self._dataframe.columns)]):
                        raise ValueError(
                            "The parameter '{param}' is not contained in this dataset.".format(
                                param=item
                            )
                        )
                    params = stage_cols | params
                elif item == "data":
                    meta = data_cols | meta
                elif item == "flags":
                    meta = flag_cols | meta
                elif re.search(param_re, item):
                    param_arg = ":" + item + ":"
                    params = self._dataframe.columns.str.contains(param_arg) | params
                    if not params.any():
                        raise ValueError(
                            "The parameter '{param}' is not contained in this dataset.".format(
                                param=item
                            )
                        )
                elif re.search(station_re, item):
                    station_arg = ":" + item + ":"
                    sites = self._dataframe.columns.str.contains(station_arg) | sites
                    if not sites.any():
                        raise ValueError(
                            "The site '{site}' is not in this dataset.".format(
                                site=item
                            )
                        )
                else:
                    raise ValueError(
                        "The argument '{item}' is not recognized.".format(item=item)
                    )
        if not sites.any():  # If no sites are selected, select them all.
            sites = all_cols
        if not params.any():  # If no params are selected, select them all.
            params = all_cols
        if (
            not meta.any()
        ):  # If neither flags nor data are selected, select data columns.
            meta = data_cols
        selection = sites & params & meta
        requested_df = self._dataframe.loc[:, selection]
        return requested_df

    def get_data(self):
        """
        .. deprecated:: version 0.2.0
           No longer needed. NWIS object will request data upon creation.
        """
        warnings.warn(
            "It is no longer necessary to call .get_data() to request data.",
            FutureWarning,
        )
        return self

    def save(self, file):
        """
        Save the dataframe and metadata to a parquet file.

        Args:
            file (str):
                the filename to save to.
        """
        extension = file.split(".")[-1]
        if extension == "parquet":
            hf.save_parquet(file, self._dataframe, self.meta)
        elif extension == "gz":
            try:
                hf.save_json_gzip(file, self.json)
            except AttributeError as err:
                print(
                    "Hydrofunctions can only save NWIS objects using gzip if the NWIS"
                    " object still has its original WaterML JSON. You might be able "
                    "to fix this problem if you call NWIS using the 'file' parameter "
                    "so that the JSON is saved immediately after the request is made."
                )
                raise err
        else:
            raise OSError(
                f"The file type extension '.{extension}' in the file name {file} is not recognized by HydroFunctions."
            )
        return self

    def read(self, file):
        """
        Read from a zipped WaterML file '.json.gz' or from a parquet file.

        Args:
            file (str):
                the filename to read from.
        """
        extension = file.split(".")[-1]
        if extension == "parquet":
            self._dataframe, self.meta = hf.read_parquet(file)
        elif extension == "gz":
            self.json = hf.read_json_gzip(file)
            self._dataframe, self.meta = hf.extract_nwis_df(self.json)
        else:
            raise OSError(
                f"The file type extension '.{extension}' in the file name {file} is not recognized by HydroFunctions."
            )
        return self



"""
hydrofunctions.hydrofunctions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the main functions used in an interactive session.

-----
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import logging
import requests
import numpy as np
import pandas as pd
import json
import gzip
import pyarrow as pa
import pyarrow.parquet as pq
from pandas.tseries.frequencies import to_offset

# Change to relative import: from . import exceptions
# https://axialcorps.com/2013/08/29/5-simple-rules-for-building-great-python-packages/
from . import exceptions
import warnings
from . import validate
from . import helpers

logger = logging.getLogger(__name__)


def select_data(nwis_df):
    """Create a boolean array of columns that contain data.

    Args:
        nwis_df:
            A pandas dataframe created by ``extract_nwis_df``.

    Returns:
        an array of Boolean values corresponding to the columns in the
        original dataframe.

    Example:

        >>> my_dataframe[:, select_data(my_dataframe)]

        returns a dataframe with only the data columns; the qualifier columns
        do not show.
    """
    data_regex = r"[0-9]$"
    return nwis_df.columns.str.contains(data_regex)


def calc_freq(index):
    # Method 0: calc_freq() was called, but we haven't done anything yet.
    method = 0
    if isinstance(index, pd.DataFrame):
        index = index.index
    try:
        # Method 1: Try the direct approach first. Maybe freq has already been set.
        freq = index.freq
        method = 1
    except AttributeError:
        # index.freq does not exist, so let's keep trying.
        freq = None

    if freq is None:
        # Method 2: Use the built-in pd.infer_freq(). It raises ValueError
        #    when it fails, so catch ValueErrors and keep trying.
        try:
            freq = to_offset(pd.infer_freq(index))
            method = 2
        except ValueError:
            pass

    if freq is None:
        # Method 3: divide the length of time by the number of observations.
        freq = (index.max() - index.min()) / len(index)
        if pd.Timedelta("13 minutes") < freq < pd.Timedelta("17 minutes"):
            freq = to_offset("15min")
        elif pd.Timedelta("27 minutes") < freq < pd.Timedelta("33 minutes"):
            freq = to_offset("30min")
        elif pd.Timedelta("55 minutes") < freq < pd.Timedelta("65 minutes"):
            freq = to_offset("60min")
        else:
            freq = None
        method = 3

    if freq is None:
        # Method 4: Subtract two adjacent values and use the difference!
        if len(index) > 3:
            freq = to_offset(abs(index[2] - index[3]))
            method = 4
            logger.debug(
                "calc_freq4:"
                + str(freq)
                + "= index[2]:"
                + str(index[3])
                + "- index [3]:"
                + str(index[2])
            )

    if freq is None:
        # Method 5: If all else fails, freq is 0 minutes!
        warnings.warn(
            "It is not possible to determine the frequency "
            "for one of the datasets in this request. "
            "This dataset will be set to a frequency of "
            "0 minutes",
            exceptions.HydroUserWarning,
        )

        freq = to_offset("0min")
        method = 5

    debug_msg = "Calc_freq method:" + str(method) + "freq:" + str(freq)
    logger.debug(debug_msg)
    return pd.Timedelta(freq)


def get_nwis(
    site,
    service="dv",
    start_date=None,
    end_date=None,
    stateCd=None,
    countyCd=None,
    bBox=None,
    parameterCd="all",
    period=None,
    verbose=True,
):
    """Request stream gauge data from the USGS NWIS.

    Args:
        site (str or list of strings):
            a valid site is '01585200' or ['01585200', '01646502']. site
            should be `None` if stateCd or countyCd are not `None`.

        service (str):
            can either be 'iv' or 'dv' for instantaneous or daily data.
                - 'dv'(default): daily values. Mean value for an entire day.
                - 'iv': instantaneous value measured at this time. Also known\
                    as 'Real-time data'. Can be measured as often as every\
                    five minutes by the USGS. 15 minutes is more typical.

        start_date (str):
           should take on the form yyyy-mm-dd

        end_date (str):
            should take on the form yyyy-mm-dd

        stateCd (str):
            a valid two-letter state postal abbreviation. Default is `None`.

        countyCd (str or list of strings):
            a valid county abbreviation. Default is `None`.

        bBox (str, list, or tuple):
            a set of coordinates that defines a bounding box.
                * Coordinates are in decimal degrees
                * Longitude values are negative (west of the prime meridian).
                * Latitude values are positive (north of the equator).
                * comma-delimited, no spaces, if provided as a string.
                * The order of the boundaries should be: "West,South,East,North"
                * Example: "-83.000000,36.500000,-81.000000,38.500000"

        parameterCd (str or list of strings):
            NWIS parameter code. Usually a five digit code. Default is 'all'.\
            A valid code can also be given as a list: ``parameterCd=['00060','00065']``
                * if value of 'all' is submitted, then NWIS will return every \
                    parameter collected at this site. (default option)
                * stage: '00065'
                * discharge: '00060'
                * not all sites collect all parameters!
                * See https://nwis.waterdata.usgs.gov/usa/nwis/pmcodes for full list

        period (str):
            NWIS period code. Default is `None`.
                * Format is "PxxD", where xx is the number of days before today.
                * Either use start_date or period, but not both.
        
        verbose (bool):
            If True (default); will print confirmation messages with the url before and 
            after the request is made.

    Returns:
        a response object. This function will always return the response,
            even if the NWIS returns a status_code that indicates a problem.

            * response.url: the url we used to request data
            * response.json: the content translated as json
            * response.status_code: the internet status code
                - '200': is a good request
                - non-200 codes will be reported as a warning.
                - '400': is a 'Bad Request'-- the parameters did not make sense
                - see <https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html> for more codes and meaning.
            * response.ok: `True` when we get a '200' status_code

    Raises:
        ConnectionError: due to connection problems like refused connection
            or DNS Error.

        SyntaxWarning: when NWIS returns a response code that is not 200.

    **Example:**

        >>> import hydrofunctions as hf
        >>> response = hf.get_nwis('01585200', 'dv', '2012-06-01', '2012-07-01')

        >>> response
        <response [200]>

        >>> response.json()
        *JSON ensues*

        >>> hf.extract_nwis_df(response)
        *a Pandas dataframe appears*

    Other Valid Ways to Make a Request::

        >>> sites = ['07180500', '03380475', '06926000'] # Request a list of sites.
        >>> service = 'iv'  # Request real-time data
        >>> days = 'P10D'  # Request the last 10 days.
        >>> stage = '00065' # Sites that collect discharge usually collect water depth too.
        >>> response2 = hf.get_nwis(sites, service, period=days, parameterCd=stage)

    Request Data By Location::

        >>> # Request the most recent daily data for every site in Maine
        >>> response3 = hf.get_nwis(None, 'dv', stateCd='ME')
        >>> response3
        <Response [200]>

    The specification for the USGS NWIS IV service is located here:
    http://waterservices.usgs.gov/rest/IV-Service.html
    """

    service = validate.check_NWIS_service(service)

    if parameterCd == "all":
        parameterCd = None

    header = {"Accept-encoding": "gzip", "max-age": "120"}

    values = {
        # specify version of nwis json. Based on WaterML1.1
        # json,1.1 works; json%2C works; json1.1 DOES NOT WORK
        "format": "json,1.1",
        "sites": validate.check_parameter_string(site, "site"),
        "stateCd": stateCd,
        "countyCd": validate.check_parameter_string(countyCd, "county"),
        "bBox": validate.check_NWIS_bBox(bBox),
        "parameterCd": validate.check_parameter_string(parameterCd, "parameterCd"),
        "period": period,
        "startDT": start_date,
        "endDT": end_date,
    }

    # Check that site selection parameters are exclusive!
    total = helpers.count_number_of_truthy([site, stateCd, countyCd, bBox])
    if total == 1:
        pass
    elif total > 1:
        raise ValueError(
            "Select sites using either site, stateCd, "
            "countyCd, or bBox, but not more than one."
        )
    elif total < 1:
        raise ValueError(
            "Select sites using at least one of the following "
            "arguments: site, stateCd, countyCd or bBox."
        )

    # Check that time parameters are not both set.
    # If neither is set, then NWIS will return the most recent observation.
    if start_date and period:
        raise ValueError(
            "Use either start_date or period, or neither, " "but not both."
        )

    if not (start_date or period):
        # User didn't specify time; must be requesting most recent data.
        # See issue #49.
        pass

    url = "https://waterservices.usgs.gov/nwis/"
    url = url + service + "/?"
    if verbose:
        print(f"Requesting data from {url}...", end="\r")
    response = requests.get(url, params=values, headers=header)
    if verbose:
        print("Requested data from", response.url)
    # requests will raise a 'ConnectionError' if the connection is refused
    # or if we are disconnected from the internet.

    # .get_nwis() will always return the response.

    # Higher-level code that calls get_nwis() may decide to handle or
    # report status codes that indicate something went wrong.

    # Issue warnings for bad status codes
    nwis_custom_status_codes(response)
    if not response.text:
        raise exceptions.HydroNoDataError(
            "The NWIS has returned an empty string for this request."
        )

    return response


def get_nwis_property(nwis_dict, key=None, remove_duplicates=False):
    """Returns a list containing property data from an NWIS response object.

    Args:
        nwis_dict (dict):
            the json returned in a response object as produced by ``get_nwis().json()``.

        key (str):
            a valid NWIS response property key. Default is `None`. The index is \
            returned if key is `None`. Valid keys are:
                * None
                * name - constructed name "provider:site:parameterCd:statistic"
                * siteName
                * siteCode
                * timeZoneInfo
                * geoLocation
                * siteType
                * siteProperty
                * variableCode
                * variableName
                * variableDescription
                * valueType
                * unit
                * options
                * noDataValue
        remove_duplicates (bool):
            a flag used to remove duplicate values in the returned list.

    Returns:
        a list with the data for the passed key string.

    Raises:
        HydroNoDataError
            when the request is valid, but NWIS has no data for \
            the parameters provided in the request.

        ValueError when the key is not available.
    """
    # nwis_dict = response_obj.json()

    # strip header and all metadata. ts is the 'timeSeries' element of the
    # response; it is an array of objects that contain time series data.
    ts = nwis_dict["value"]["timeSeries"]
    msg = "The NWIS reports that it does not have any data for this request."

    if len(ts) < 1:
        raise exceptions.HydroNoDataError(msg)

    # This predefines what to expect in the response.
    # Would it be better to look in the response for the key?
    # Pseudo code
    # skip stations with no data
    # if key in tts['variable']:
    #    v = etc
    # elif key in tts['sourceInfo']:
    #    v = etc
    # elif key in tts:
    #    v = etc
    # else just return index or raise an error later
    #
    sourceInfo = [
        "siteName",
        "siteCode",
        "timeZoneInfo",
        "geoLocation",
        "siteType",
        "siteProperty",
    ]
    variable = [
        "variableCode",
        "variableName",
        "variableDescription",
        "valueType",
        "unit",
        "options",
        "noDataValue",
    ]
    root = ["name"]
    vals = []
    try:
        for idx, tts in enumerate(ts):
            d = tts["values"][0]["value"]
            # skip stations with no data
            if len(d) < 1:
                continue
            if key in variable:
                v = tts["variable"][key]
            elif key in sourceInfo:
                v = tts["sourceInfo"][key]
            elif key in root:
                v = tts[key]
            else:
                v = idx  # just return index
            if remove_duplicates:
                if v not in vals:
                    vals.append(v)
            else:
                vals.append(v)
    # Why catch this? If we can't find the key, we already return the index.
    except:  # TODO: dangerous to use bare 'except'  clauses.
        msg = 'The selected key "{}" could not be found'.format(key)
        raise ValueError(msg)
    return vals


def extract_nwis_df(nwis_dict, interpolate=True):
    """Returns a Pandas dataframe and a metadata dict from the NWIS response
    object or the json dict of the response.

    Args:
        nwis_dict (obj):
            the json from a response object as returned by get_nwis().json().
            Alternatively, you may supply the response object itself.

    Returns:
        a pandas dataframe.

    Raises:
        HydroNoDataError
            when the request is valid, but NWIS has no data for
            the parameters provided in the request.

        HydroUserWarning
            when one dataset is sampled at a lower frequency than
            another dataset in the same request.
    """
    if type(nwis_dict) is not dict:
        nwis_dict = nwis_dict.json()

    # strip header and all metadata.
    ts = nwis_dict["value"]["timeSeries"]
    if ts == []:
        # raise a HydroNoDataError if NWIS returns an empty set.
        #
        # Ideally, an empty set exception would be raised when the request
        # is first returned, but I do it here so that the data doesn't get
        # extracted twice.
        # TODO: raise this exception earlier??
        #
        # ** Interactive sessions should have an error raised.
        #
        # **Automated systems should catch these errors and deal with them.
        # In this case, if NWIS returns an empty set, then the request
        # needs to be reconsidered. The request was valid somehow, but
        # there is no data being collected.

        raise exceptions.HydroNoDataError(
            "The NWIS reports that it does not " "have any data for this request."
        )

    # create a list of time series;
    # set the index, set the data types, replace NaNs, sort, find the first and last

    collection = []
    starts = []
    ends = []
    freqs = []
    meta = {}
    for series in ts:
        full_series_name = series["name"]
        name_list = full_series_name.split(":")
        agency = str(name_list[0])
        site_id = agency + ":" + str(name_list[1])
        parameter_cd = str(name_list[2])
        stat = str(name_list[3])
        siteName = series["sourceInfo"]["siteName"]
        siteLatLongSrs = series["sourceInfo"]["geoLocation"]["geogLocation"]
        noDataValues = series["variable"]["noDataValue"]
        variableDescription = series["variable"]["variableDescription"]
        unit = series["variable"]["unit"]["unitCode"]
        values = series["values"]
        for method in values:
            data = method["value"]
            # This line assumes only one method per parameter. See issue #77.
            # data = series["values"][0]["value"]
            if data == []:
                # This parameter has no data. Skip to next series.
                continue
            if len(data) == 1:
                # This parameter only contains the most recent reading.
                # See Issue #49
                pass
            method_description = method["method"][0]["methodDescription"]
            method_id = str(method["method"][0]["methodID"])
            # use method_mod as a modifier for altering parameter names.
            method_mod = "-" + method_id
            if len(values) == 1:
                # If there is only one method, don't bother recording method #.
                method_mod = ""
            series_name = site_id + ":" + parameter_cd + method_mod + ":" + stat
            qualifiers_name = series_name + "_qualifiers"
            DF = pd.DataFrame(data=data)
            DF.index = pd.to_datetime(DF.pop("dateTime"), utc=True)
            DF["value"] = DF["value"].astype(float)
            DF = DF.replace(to_replace=noDataValues, value=np.nan)
            DF["qualifiers"] = DF["qualifiers"].apply(lambda x: ",".join(x))
            DF.rename(
                columns={"qualifiers": qualifiers_name, "value": series_name},
                inplace=True,
            )
            DF.sort_index(inplace=True)
            local_start = DF.index.min()
            local_end = DF.index.max()
            starts.append(local_start)
            ends.append(local_end)
            local_freq = calc_freq(DF.index)
            freqs.append(local_freq)
            if not DF.index.is_unique:
                print(
                    "Series index for "
                    + series_name
                    + " is not unique. Attempting to drop identical rows."
                )
                DF = DF.drop_duplicates(keep="first")
                if not DF.index.is_unique:
                    print(
                        "Series index for "
                        + series_name
                        + " is STILL not unique. Dropping first rows with duplicated date."
                    )
                    DF = DF[~DF.index.duplicated(keep="first")]
            if local_freq > pd.Timedelta(to_offset("0min")):
                local_clean_index = pd.date_range(
                    start=local_start, end=local_end, freq=local_freq, tz="UTC"
                )
                # if len(local_clean_index) != len(DF):
                # This condition happens quite frequently with missing data.
                # print(str(series_name) + "-- clean index length: "+ str(len(local_clean_index)) + " Series length: " + str(len(DF)))
                DF = DF.reindex(index=local_clean_index, copy=True)
            else:
                # The dataframe DF must contain only the most recent data.
                pass
            qual_cols = DF.columns.str.contains("_qualifiers")
            # https://stackoverflow.com/questions/21998354/pandas-wont-fillna-inplace
            # Instead, create a temporary dataframe, fillna, then copy back into original.
            DFquals = DF.loc[:, qual_cols].fillna("hf.missing")
            DF.loc[:, qual_cols] = DFquals

            if local_freq > pd.Timedelta(to_offset("0min")):
                variableFreq_str = str(to_offset(local_freq))
            else:
                variableFreq_str = str(to_offset("0min"))
            parameter_info = {
                "variableFreq": variableFreq_str,
                "variableUnit": unit,
                "variableDescription": variableDescription,
                "methodID": method_id,
                "methodDescription": method_description,
            }
            site_info = {
                "siteName": siteName,
                "siteLatLongSrs": siteLatLongSrs,
                "timeSeries": {},
            }
            # if site is not in meta keys, add it.
            if site_id not in meta:
                meta[site_id] = site_info
            # Add the variable info to the site dict.
            meta[site_id]["timeSeries"][parameter_cd + method_mod] = parameter_info
            collection.append(DF)

    if len(collection) < 1:
        # It seems like this condition should not occur. The NWIS trims the
        # response and returns an empty nwis_dict['value']['timeSeries']
        # if none of the parameters requested have data.
        # If at least one of the paramters have data,
        # then the empty series will get delivered, but with no data.
        # Compare these requests:
        # empty:               https://nwis.waterservices.usgs.gov/nwis/iv/?format=json&sites=01570500&startDT=2018-06-01&endDT=2018-06-01&parameterCd=00045
        # one empty, one full: https://nwis.waterservices.usgs.gov/nwis/iv/?format=json&sites=01570500&startDT=2018-06-01&endDT=2018-06-01&parameterCd=00045,00060
        raise exceptions.HydroNoDataError(
            "The NWIS does not have any data for"
            " the requested combination of sites"
            ", parameters, and dates."
        )
    startmin = min(starts)
    endmax = max(ends)
    # Remove all frequencies of zero from freqs list.
    zero = pd.Timedelta("0min")
    freqs_no_zeros = list(filter(lambda x: x > zero, freqs))
    if len(freqs_no_zeros) > 0:
        freqmin = min(freqs_no_zeros)
        freqmax = max(freqs_no_zeros)
        if freqmin != freqmax:
            warnings.warn(
                "One or more datasets in this request is going to be "
                "'upsampled' to " + str(freqmin) + " because the data "
                "were collected at a lower frequency of " + str(freqmax),
                exceptions.HydroUserWarning,
            )
        clean_index = pd.date_range(start=startmin, end=endmax, freq=freqmin, tz="UTC")
        cleanDF = pd.DataFrame(index=clean_index)
        for dataset in collection:
            cleanDF = pd.concat([cleanDF, dataset], axis=1)
        # Replace lines with missing _qualifier flags with hf.upsampled
        qual_cols = cleanDF.columns.str.contains("_qualifiers")
        cleanDFquals = cleanDF.loc[:, qual_cols].fillna("hf.upsampled")
        cleanDF.loc[:, qual_cols] = cleanDFquals
        if interpolate:
            # TODO: mark interpolated values with 'hf.interp'
            # select data, then replace Nans with interpolated values.
            data_cols = cleanDF.columns.str.contains(r"[0-9]$")
            cleanDFdata = cleanDF.loc[:, data_cols].interpolate()
            cleanDF.loc[:, data_cols] = cleanDFdata
    else:
        # If datasets only contain most recent data, then
        # don't set an index or a freq. Just concat all of the datasets.
        # Alternatively, to solve issue #54 (Requests for only the most recent
        # data should be parsed differently) We could combine the different dataframes
        # in collection using a different procedure.
        cleanDF = pd.concat(collection, axis=1)

    cleanDF.index.name = "datetimeUTC"

    if not DF.index.is_unique:
        DF = DF[~DF.index.duplicated(keep="first")]
    if not DF.index.is_monotonic:
        DF.sort_index(axis=0, inplace=True)

    return cleanDF, meta


def nwis_custom_status_codes(response):
    """
    Raise custom warning messages from the NWIS when it returns a
    status_code that is not 200.

    Args:
        response: a response object as returned by get_nwis().

    Returns:
        `None` if response.status_code == 200

    Raises:
        HydroNoDataError: when a non-200 status code is returned.
            https://en.wikipedia.org/wiki/List_of_HTTP_status_codes

    Note:
        NWIS status_code messages come from:
            https://waterservices.usgs.gov/docs/portable_code.html
        Additional status code documentation:
            https://waterservices.usgs.gov/rest/IV-Service.html#Error
    """
    nwis_msg = {
        "200": "OK",
        "400": "400 Bad Request - "
        "This often occurs if the URL arguments "
        "are inconsistent. For example, if you submit a request using "
        "a startDT and an endDT with the period argument. "
        "An accompanying error should describe why the request was "
        "bad." + "\nError message from NWIS: {}".format(response.reason),
        "403": "403 Access Forbidden - "
        "This should only occur if for some reason the USGS has "
        "blocked your Internet Protocol (IP) address from using "
        "the service. This can happen if we believe that your use "
        "of the service is so excessive that it is seriously "
        "impacting others using the service. To get unblocked, "
        "send us the URL you are using along with the IP using "
        "this form. We may require changes to your query and "
        "frequency of use in order to give you access to the "
        "service again.",
        "404": "404 Not Found - "
        "Returned if and only if the query expresses a combination "
        "of elements where data do not exist. For multi-site "
        "queries, if any data are found, it is returned for those "
        "site/parameters/date ranges where there are data.",
        "503": "500 Internal Server Error - "
        "If you see this, it means there is a problem with the web "
        "service itself. It usually means the application server "
        "is down unexpectedly. This could be caused by a host of "
        "conditions, but changing your query will not solve this "
        "problem. The NWIS application support team has to fix it. Most "
        "of these errors are quickly detected and the support team "
        "is notified if they occur.",
    }
    if response.status_code == 200:
        return None
    # All other status codes will raise an exception.
    # Use the status_code as a key, return None if key not in dict
    msg = (
        "The NWIS returned a code of {}.\n".format(response.status_code)
        + nwis_msg.get(str(response.status_code))
        + "\nURL used in this request: {}".format(response.url)
    )
    raise exceptions.HydroNoDataError(msg)


def read_parquet(filename):
    """Read a hydrofunctions parquet file.

    This function will read a parquet file that was saved by
    hydrofunctions.save_parquet() and return a dataframe and a metadata dictionary.

    Args:
        filename (str): A string with the filename and extension.

    Returns:
        dataframe (pd.DataFrame): a pandas dataframe.
        meta (dict): a dictionary with the metadata for the NWIS data request, if it exists.
    """
    pa_table = pq.read_table(filename)
    dataframe = pa_table.to_pandas()
    dataframe.index.freq = calc_freq(dataframe.index)
    meta_dict = pa_table.schema.metadata
    if b"hydrofunctions_meta" in meta_dict:
        meta_string = meta_dict[b"hydrofunctions_meta"].decode()
        meta = json.loads(meta_string)
    else:
        meta = None
    return dataframe, meta


def save_parquet(filename, dataframe, hf_meta):
    """Save a hydrofunctions parquet file.

    This function will save a dataframe and a dictionary into the parquet format.
    Parquet files are a compact, easy to process format that work well with Pandas and
    large datasets. This function will accompany the dataframe with a dictionary of NWIS
    metadata that is produced by the hydrofunctions.extract_nwis_df() function. This
    file can then be read by the hydrofunctions.read_parquet() function.

    Args:
        filename (str): A string with the filename and extension.
        dataframe (pd.DataFrame): a pandas dataframe.
        hf_meta (dict): a dictionary with the metadata for the NWIS data request, if it exists.
    """
    if len(filename.split(".")) == 1:
        filename = filename + ".gz.parquet"

    table = pa.Table.from_pandas(dataframe, preserve_index=True)
    meta_dict = table.schema.metadata
    hf_string = json.dumps(hf_meta).encode()
    meta_dict[b"hydrofunctions_meta"] = hf_string
    table = table.replace_schema_metadata(meta_dict)
    pq.write_table(table, filename, compression="gzip")


def read_json_gzip(filename):
    """Read a gzipped JSON file into a Python dictionary.

    Reads JSON files that have been zipped and returns a Python dictionary.
    Usually the files should have an extension .json.gz
    Hydrofunctions uses this function to store the original JSON format WaterML
    response from the USGS NWIS.

    Args:
        filename (str): A string with the filename and extension.

    Returns:
        a dictionary of the file contents.
    """
    with gzip.open(filename, "rb") as zip_file:
        zip_dict = json.loads(zip_file.read())
        return zip_dict


def save_json_gzip(filename, json_dict):
    """Save a Python dictionary as a gzipped JSON file.

    This save function is especially designed to compress and save the original
    JSON response from the USGS NWIS. If no file extension is specified, then a
    .json.gz extension will be provided.

    Args:
        filename (str): A string with the filename and extension.
        json_dict (dict): A dictionary representing the json content.
    """
    if len(filename.split(".")) == 1:
        filename = filename + "json.gz"

    with gzip.open(filename, "wt", encoding="ascii") as zip_file:
        json.dump(json_dict, zip_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import re

with open("pypi_readme.rst") as readme_file:
    # with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("docs/history.rst") as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.read()


def relative2absolute(input, old, new):
    """Replaces every instance of rel_key in input with absolute_stem.
    Use this to change relative links to absolute links in pypi.
    """
    pattern = old
    p = re.compile(pattern)
    output = p.sub(new, input)
    return output


relative = r"_static"
stem = r"https://raw.githubusercontent.com/mroberge/hydrofunctions/master/_static"
# readme = relative2absolute(readme, relative, stem)

setup(
    name="hydrofunctions",
    version="0.2.2",
    description="A suite of convenience functions for exploring water data in a Jupyter Notebook.",
    long_description=readme,  # + "\n\n" + history,
    long_description_content_type="text/x-rst",
    author="Martin Roberge",
    author_email="mroberge@towson.edu",
    url="https://github.com/mroberge/hydrofunctions",
    packages=find_packages(),
    package_dir={"hydrofunctions": "hydrofunctions"},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="hydrofunctions hydrology USGS stream gauge water NWIS",
    project_urls={
        "Documentation": "https://hydrofunctions.readthedocs.io",
        "Source": "https://github.com/mroberge/hydrofunctions",
        "Latest": "https://github.com/mroberge/hydrofunctions/tree/develop",
        "Tracker": "https://github.com/mroberge/hydrofunctions/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Hydrology",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Utilities",
    ],
    test_suite="tests",
)



"""
hydrofunctions.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains all of the custom exceptions defined in this package. The
base class is HydroException and all custom exceptions are subclasses of
HydroException.

Use the errors like this::

    try:
        #some code here that might return no data
        #more code that might get encoded improperly
    except HydroNoDataError('This site has no data'):
        # handle error here.
    except HydroEncodeError():
        # handle this error here.
    else:
        # code to complete if there is no exception raised.
    finally:
        # code that you want to run whether an exception is raised or not.
        # If an exception wasn't caught, then this code gets run, and the
        # exception gets re-raised after this finally clause gets run.

Keep the try clause short: if you put too many things in there, it can be
difficult to figure out what broke. On the other hand, like in my example
above, it is more readable if you group a series of statements and then
handle their exceptions together.

**Example:**

    >>> raise HydroNoDataError("Oh no, NWIS doesn't have this data for you!")

https://axialcorps.com/2013/08/29/5-simple-rules-for-building-great-python-packages/

-----
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import logging
import warnings

logger = logging.getLogger(__name__)


class HydroException(Exception):
    """This is the base class for all exceptions created for the
    HydroFunctions package. This class is not meant to be raised.
    """

    def __init__(self, msg=""):
        self.msg = msg
        logger.error(msg)  # This message will be logged if logging has been started.

    def __str__(self):
        return self.msg

    pass


class HydroNoDataError(HydroException):
    """Raised when a service returns an empty dataset or indicates that\
        it has no data for the request.

        **Usage**::

            raise HydroNoDataError("The NWIS service had no data for this request.")

        Do not catch this error for interactive sessions: The user should
        get a useful message from the error when they try to request something
        that doesn't exist.

        Catch this error in automated systems so that the system can reconsider
        the request and either fix the request or move on to the next
        request.

        **Example**::

            try:
                hf.NWIS('666666666')
            except HydroNoDataError as err:
                print("This is just to illustrate how to capture this error.")
                print(err)
    """

    pass


class HydroEncodeError(HydroException):
    """Raised when an error occurs while encoding or decoding an argument.

    **Example**::

        try:
            # bunch of code from your package
        except HydroException:
            # blanked condition to handle all errors from your package
    """

    pass


class HydroUserWarning(UserWarning):
    """Warn user of a hazardous condition or when an action has been triggered\
        that may be unexpected.

        This is the base class for all warnings created for the HydroFunctions
        package. This class can be used if there is no more specific warning
        available.

        **Usage**::

            import hydrofunctions as hf
            import warnings
            ... code
            warnings.warn('This is my warning message.', hf.HydroUserWarning)

        Note:
            Warnings can be hidden or turned off depending on how the user is
            accessing Python and the settings for their interface.

            Use HydroException if a process must be shut down, or is doomed to
            fail anyway. This will at least give the user a helpful error
            message.
    """

    def __init__(self, msg=""):
        self.msg = msg
        logger.error(msg)  # This message will be logged if logging has been started.

    def __str__(self):
        return self.msg

    pass



"""
hydrofunctions.validate
~~~~~~~~~~~~~~~~~~~~~~~

This module contains functions for testing that user input is valid.

Why 'pre-check' user imputs, instead of using standard
python duck typing? These functions are meant to enhance an interactive
session for the user, and will check a user's parameters
before requesting data from an online resource. Otherwise, the server will
return a 404 code and the user will have no idea why. Hydrofunctions tries to raise
an exception (usually a TypeError) before a request is made, so that the user
can fix their request. It also tries to provide a helpful error message to an
interactive session user.

Suggested format for these functions:

* first check that the input is a string,
* then do a regular expression to check that the input is more or less valid.
* raise exceptions when user input breaks format.

-----
"""
from __future__ import absolute_import, print_function, division, unicode_literals
import re


def check_parameter_string(candidate, param):
    """Checks that a parameter is a string or a list of strings."""
    parameters = {
        "site": "NWIS station id(s) should be a string or list of strings,"
        + "often in the form of an eight digit number enclosed in quotes.",
        "parameterCd": "NWIS parameter codes are five-digit strings that specify "
        + "the parameter that is being measured at the site. Common "
        + "codes are '00060' for stream stage in feet, '00065' for "
        + "stream discharge in cubic feet per second, and '72019' for "
        + "groundwater levels. Not all sites collect data for all "
        + "parameters. See a complete list of physical parameters here: "
        + "https://help.waterdata.usgs.gov/parameter_cd?group_cd=PHY "
        + "You may request multiple parameters by submitting a comma-"
        + "delimited string of codes with no spaces, or by submitting "
        + "a list of codes, like this: parameterCd = '00065,00060' or "
        + "parameterCd = ['00065', '00060'] ",
        "county": "The NWIS county parameter accepts a five-digit string or "
        + "a list of five-digit strings to select all of the sites "
        + "within a county or list of counties. "
        + "Example: '51059' or ['51059', '51061'] are acceptable.",
        "state": "This parameter uses US two-letter postal codes "
        + "such as 'MD' for Maryland or 'AZ' for Arizona.",
        "default": "This parameter should be a string or a list of strings.",
    }
    if param in parameters:
        msg = parameters[param] + " Actual value: {}".format(candidate)
    else:
        msg = (
            "This parameter should be a string or a list of strings."
            + " Actual value: {}".format(candidate)
        )

    if candidate is None:
        return None
    elif isinstance(candidate, str) and candidate:
        return candidate
    elif isinstance(candidate, (list, tuple)) and candidate:
        for s in candidate:
            if not isinstance(s, str):
                raise TypeError(msg + "   bad element: {}".format(s))
        return ",".join(str(s) for s in candidate)
    else:
        raise TypeError(msg)


def check_NWIS_bBox(input):
    """Checks that the USGS bBox is valid."""
    msg = (
        "NWIS bBox should be a string, list of strings, or tuple "
        + "containing the longitude and latitude of the lower left corner "
        + "of the bounding box, followed by the longitude and latitude "
        + "of the upper right corner of the bounding box. Most often in "
        + 'the form of "ll_long,ll_lat,ur_long,ur_lat" . '
        + "All latitude and longitude values should have less than 8 "
        + "places. "
        + "Actual value: {}".format(input)
    )
    if input is None:
        return None
    # assume that if it is a string it will be fine as is.
    # don't accept a series of sites in a single string.
    # Test for and reject empty strings: empty strings are false.
    if isinstance(input, str) and input:
        t = input.split(",")
        if len(t) < 4:
            raise TypeError(msg)
        return input
    # test for input is a list and it is not empty
    elif (isinstance(input, list) or isinstance(input, tuple)) and input:
        if len(input) < 4:
            raise TypeError(msg)
        # format:  [-83.000000, 36.500000, -81.000000, 38.500000] ==> '-83.000000,36.500000,-81.000000,38.500000'
        return ",".join([str(s) for s in input])
    else:
        raise TypeError(msg)


def check_NWIS_service(input):
    """Checks that the service is valid: either 'iv' or 'dv'"""
    if input is None:
        return None
    if input in ["iv", "dv"]:
        return input
    else:
        raise TypeError(
            "The NWIS service type accepts 'dv' for daily values, "
            "or 'iv' for instantaneous values. Actual value: "
            "{}".format(input)
        )


def check_datestr(input):
    """Checks that the start_date or end_date parameter is in yyyy-mm-dd format."""
    # Use a regular expression to ensure in form of yyyy-mm-dd
    if input is None:
        return None
    pattern = r"[1-2]\d\d\d-[0-1]\d-[0-3]\d\Z"
    datestr = re.compile(pattern)
    if isinstance(input, str) and datestr.match(input):
        return input
    else:
        raise TypeError(
            "Dates should be a string in the form of 'YYYY-MM-DD' "
            "enclosed in quotes. Actual value: {}".format(input)
        )


def check_period(input):
    """Checks that the period parameter in is the P##D format, where ## is
    the number of days before now.
    """
    if input is None:
        return None
    # TODO: check how many days maximum NWIS is willing to respond to.
    # This pattern sets a maximum of 999 days (between 1 and 3 digits).
    pattern = r"^P\d{1,3}D$"
    periodstr = re.compile(pattern)
    if isinstance(input, str) and periodstr.match(input):
        return input
    else:
        raise TypeError(
            "Period should be a string in the form of 'PxD', "
            "where x represents the number of days before today, "
            "with a maximum of 999 days. "
            "Example: to request the previous 10 days, "
            "enter 'period=P10D'. Actual value entered: {}".format(input)
        )
